# -*- coding: utf-8 -*-
"""
Created on Wed Jan 10 16:54:14 2018

@author: Weber Sébastien
"""
from __future__ import annotations
from importlib import import_module
from collections import OrderedDict
import copy
import os
from pathlib import Path
import sys
from typing import List, Tuple, Union
import time

from easydict import EasyDict as edict
import numpy as np
from qtpy import QtWidgets
from qtpy.QtCore import Qt, QObject, Slot, QThread, Signal

from pymodaq.utils.data import DataFromPlugins, DataToExport, Axis, DataDistribution
from pymodaq.utils.logger import set_logger, get_module_name
from pymodaq.control_modules.utils import ControlModule
from pymodaq.utils.gui_utils.file_io import select_file
from pymodaq.utils.tcp_ip.tcp_server_client import TCPClient
from pymodaq.utils.gui_utils.widgets.lcd import LCD
from pymodaq.utils.config import Config, get_set_local_dir
from pymodaq.utils.h5modules.browsing import browse_data
from pymodaq.utils.h5modules.saving import H5Saver
from pymodaq.utils.h5modules import module_saving
from pymodaq.utils.h5modules.backends import Node
from pymodaq.utils.daq_utils import ThreadCommand
from pymodaq.utils.parameter import ioxml
from pymodaq.utils.parameter import utils as putils
from pymodaq.control_modules.viewer_utility_classes import params as daq_viewer_params
from pymodaq.utils import daq_utils as utils
from pymodaq.utils.messenger import deprecation_msg
from pymodaq.utils.gui_utils import DockArea, get_splash_sc, Dock
from pymodaq.utils.managers.parameter_manager import ParameterManager, Parameter
from pymodaq.control_modules.daq_viewer_ui import DAQ_Viewer_UI
from pymodaq.control_modules.utils import DET_TYPES, get_viewer_plugins, DAQTypesEnum, DetectorError
from pymodaq.utils.plotting.data_viewers.viewer import ViewerBase, ViewersEnum
from pymodaq.utils.enums import enum_checker
from pymodaq.control_modules.viewer_utility_classes import DAQ_Viewer_base

logger = set_logger(get_module_name(__file__))
config = Config()

local_path = get_set_local_dir()


class DAQ_Viewer(ParameterManager, ControlModule):
    """ Main PyMoDAQ class to drive detectors

    Qt object and generic UI to drive actuators. The class is giving you full functionality to select (daq_detector),
    initialize detectors (init_hardware), grab or snap data (grab_data) and save them (save_new, save_current). If
    a DockArea is given as parent widget, the full User Interface (DAQ_Viewer_UI) is loaded allowing easy control of the
    instrument.

    Attributes
    ----------
    grab_done_signal: Signal[DataToExport]
        Signal emitted when the data from the plugin (and eventually from the data viewers) has been received. To be
        used by connected objects.
    custom_sig: Signal[ThreadCommand]
        use this to propagate info/data coming from the hardware plugin to another object
    overshoot_signal: Signal[bool]
        This signal is emitted when some 0D data from the plugin is higher than the overshoot threshold set in the
        settings

    See Also
    --------
    ControlModule, DAQ_Viewer_UI, ParameterManager

    Notes
    -----
    A particular signal from the 2D DataViewer is directly connected to the plugin: ROI_select_signal. The position and
    size of the corresponding ROI is then directly transferred to a plugin function named `ROISelect` that you have to
    create if one want to receive infos from the ROI
    """
    settings_name = 'daq_viewer_settings'
    custom_sig = Signal(ThreadCommand)  # particular case where DAQ_Viewer  is used for a custom module

    grab_done_signal = Signal(DataToExport)

    _update_settings_signal = Signal(edict)
    overshoot_signal = Signal(bool)
    data_saved = Signal()
    grab_status = Signal(bool)

    params = daq_viewer_params

    def __init__(self, parent=None, title="Testing",
                 daq_type=config('viewer', 'daq_type'),
                 dock_settings=None, dock_viewer=None):

        self.logger = set_logger(f'{logger.name}.{title}')
        self.logger.info(f'Initializing DAQ_Viewer: {title}')

        ParameterManager.__init__(self, action_list = ('save','update'))
        ControlModule.__init__(self)

        daq_type = enum_checker(DAQTypesEnum, daq_type)
        self._daq_type: DAQTypesEnum = daq_type

        self._viewer_types: List[ViewersEnum] = []
        self._viewers: List[ViewerBase] = []

        if isinstance(parent, DockArea):
            self.dockarea = parent
        else:
            self.dockarea = None

        self.parent = parent
        if parent is not None:
            self.ui: DAQ_Viewer_UI = DAQ_Viewer_UI(parent, title, daq_type=daq_type,
                                                   dock_settings=dock_settings,
                                                   dock_viewer=dock_viewer)
        else:
            self.ui: DAQ_Viewer_UI = None

        if self.ui is not None:
            QtWidgets.QApplication.processEvents()
            self.ui.add_setting_tree(self.settings_tree)
            self.ui.command_sig.connect(self.process_ui_cmds)
            self.viewers = self.ui.viewers
            self._viewer_types = self.ui.viewer_types

        self.splash_sc = get_splash_sc()

        self._title = title

        self.module_and_data_saver: Union[module_saving.DetectorSaver,
                                          module_saving.DetectorEnlargeableSaver,
                                          module_saving.DetectorExtendedSaver] = None
        self._h5saver_continuous: H5Saver = None
        self._ind_continuous_grab = 0
        self.setup_continuous_saving()

        self.settings.child('main_settings', 'DAQ_type').setValue(self.daq_type.name)
        self._detectors: List[str] = [det_dict['name'] for det_dict in DET_TYPES[self.daq_type.name]]
        if len(self._detectors) > 0:  # will be 0 if no valid plugins are installed
            self._detector: str = self._detectors[0]
        else:
            raise DetectorError('No detected Detector')
        self.settings.child('main_settings', 'detector_type').setValue(self._detector)

        self._grabing: bool = False
        self._do_bkg: bool = False
        self._take_bkg: bool = False

        self._grab_done: bool = False
        self._start_grab_time: float = 0.  # used for the refreshing rate
        self._received_data: int = 0

        self._lcd: LCD = None

        self._bkg: DataToExport = None  # buffer to store background

        self._save_file_pathname: Path = None  # to store last active path, will be an Path object
        
        self._snapshot_pathname: Path = None
        self._data_to_save_export: DataToExport = None

        self._do_save_data: bool = False

        self._set_setting_tree()  # to activate parameters of default Mock detector

        self.grab_done_signal.connect(self._save_export_data)
        self.update_plugin_config()

    def __repr__(self):
        return f'{self.__class__.__name__}: {self.title} ({self.daq_type}/{self.detector}'

    def setup_continuous_saving(self):
        """Configure the objects dealing with the continuous saving mode"""
        self.module_and_data_saver = module_saving.DetectorSaver(self)
        self._h5saver_continuous = H5Saver(save_type='detector')
        self._h5saver_continuous.show_settings(False)
        self._h5saver_continuous.settings.child('do_save').sigValueChanged.connect(self._init_continuous_save)
        if self.ui is not None:
            self.ui.add_setting_tree(self._h5saver_continuous.settings_tree)

    def process_ui_cmds(self, cmd: utils.ThreadCommand):
        """Process commands sent by actions done in the ui

        Parameters
        ----------
        cmd: ThreadCommand
            Possible values are:
                * init
                * quit
                * grab
                * snap
                * stop
                * show_log
                * detector_changed
                * daq_type_changed
                * save_current
                * save_new
                * do_bkg
                * take_bkg
                * viewers_changed
                * show_config
        """

        if cmd.command == 'init':
            self.init_hardware(cmd.attribute[0])
        elif cmd.command == 'quit':
            self.quit_fun()
        elif cmd.command == 'stop':
            self.stop()
        elif cmd.command == 'show_log':
            self.show_log()
        elif cmd.command == 'grab':
            self.grab_data(cmd.attribute, snap_state=False)
        elif cmd.command == 'snap':
            self.grab_data(False, snap_state=True)
        elif cmd.command == 'save_new':
            self.save_new()
        elif cmd.command == 'save_current':
            self.save_current()
        elif cmd.command == 'open':
            self.load_data()
        elif cmd.command == 'detector_changed':
            if cmd.attribute != '':
                self.detector_changed_from_ui(cmd.attribute)
        elif cmd.command == 'daq_type_changed':
            if cmd.attribute != '':
                self.daq_type_changed_from_ui(cmd.attribute)
        elif cmd.command == 'take_bkg':
            self.take_bkg()
        elif cmd.command == 'do_bkg':
            self.do_bkg = cmd.attribute
        elif cmd.command == 'viewers_changed':
            self._viewer_types: List[ViewersEnum] = cmd.attribute['viewer_types']
            self.viewers = cmd.attribute['viewers']
        elif cmd.command == 'show_config':
            self.config = self.show_config(self.config)
            self.ui.config = self.config

    @property
    def bkg(self) -> DataToExport:
        """Get the background data object"""
        return self._bkg

    @property
    def viewer_docks(self) -> List[Dock]:
        """list of Viewer Docks from the UI"""
        if self.ui is not None:
            return self.ui.viewer_docks

    @property
    def viewers_docks(self) -> List[Dock]:
        """list of Viewer Docks from the UI, for back compatibility"""
        deprecation_msg('viewers_docks is a deprecated property use viewer_docks instead')
        return self.viewer_docks

    def daq_type_changed_from_ui(self, daq_type: DAQTypesEnum):
        """ Apply changes from the selection of a different DAQTypesEnum in the UI

        Parameters
        ----------
        daq_type: DAQTypesEnum
        """
        daq_type = enum_checker(DAQTypesEnum, daq_type)
        self._daq_type = daq_type
        self.settings.child('main_settings', 'DAQ_type').setValue(daq_type.name)
        self.detectors = [det_dict['name'] for det_dict in DET_TYPES[daq_type.name]]
        self.detector = self.detectors[0]

    @property
    def daq_type(self) -> DAQTypesEnum:
        """Get/Set the daq_type as a DAQTypesEnum

        Update the detector property with the list of available detectors of a given daq_type
        """
        return self._daq_type

    @daq_type.setter
    def daq_type(self, daq_type: DAQTypesEnum):
        daq_type = enum_checker(DAQTypesEnum, daq_type)

        self._daq_type = daq_type
        if self.ui is not None:
            self.ui.daq_type = daq_type
        self.settings.child('main_settings', 'DAQ_type').setValue(daq_type.name)
        self.detectors = [det_dict['name'] for det_dict in DET_TYPES[daq_type.name]]
        self.detector = self.detectors[0]

    @property
    def daq_types(self) -> List[str]:
        """List of available DAQ_TYPES as keys of the DAQTypesEnum"""
        return DAQTypesEnum.names()

    def detector_changed_from_ui(self, detector: str):
        self._detector = detector
        self.update_plugin_config()
        self._set_setting_tree()

    @property
    def detector(self) -> str:
        """:obj:`str`: Get/Set the currently selected detector among available detectors"""
        return self._detector

    @detector.setter
    def detector(self, det: str):
        if det not in self.detectors:
            raise ValueError(f'{det} is not a valid Detector: {self.detectors}')
        self._detector = det
        self.update_plugin_config()
        if self.ui is not None:
            self.ui.detector = det
        self._set_setting_tree()

    @property
    def Naverage(self):
        return self.settings['main_settings', 'Naverage']

    @Naverage.setter
    def Naverage(self, ngrab: int):
        if ngrab >= 1:
            self.settings.child('main_settings', 'Naverage').setValue(ngrab)

    def update_plugin_config(self):
        parent_module = utils.find_dict_in_list_from_key_val(DET_TYPES[self.daq_type.name], 'name', self.detector)
        mod = import_module(parent_module['module'].__package__.split('.')[0])
        if hasattr(mod, 'config'):
            self.plugin_config = mod.config

    def detectors_changed_from_ui(self, detectors: List[str]):
        self._detectors = detectors

    @property
    def detectors(self) -> str:
        """:obj:`list` of :obj:`str`: List of available detectors of the current daq_type (DAQTypesEnum)"""
        return self._detectors

    @detectors.setter
    def detectors(self, detectors):
        self._detectors = detectors
        if self.ui is not None:
            self.ui.detectors = detectors

    @property
    def grab_state(self):
        """:obj:`bool`: Get the current grabbing status"""
        return self._grabing

    @property
    def do_bkg(self) -> bool:
        """:obj:`bool`: Get/Set if background subtraction should be done"""
        return self._do_bkg

    @do_bkg.setter
    def do_bkg(self, doit: bool):
        self._do_bkg = doit

    @property
    def viewers(self) -> List[ViewerBase]:
        """:obj:`list`: Get/Set the Viewers (instances of real implementation of ViewerBase class) from the UI"""
        if self.ui is not None:
            return self._viewers

    @viewers.setter
    def viewers(self, viewers):
        for viewer in self._viewers:
            try:
                viewer.data_to_export_signal.disconnect()
            except:
                pass
        for viewer in viewers:
            viewer.data_to_export_signal.connect(self._get_data_from_viewer)
            if hasattr(viewer, 'ROI_select_signal'):
                viewer.ROI_select_signal.connect(
                    lambda roi_pos_size: self.command_hardware.emit(ThreadCommand('ROISelect', roi_pos_size)))
        self._viewers = viewers

    def quit_fun(self):
        """ Quit the application, closing the hardware and other modules """

        # insert anything that needs to be closed before leaving

        if self._initialized_state:  # means  initialized
            self.init_hardware(False)
        self.quit_signal.emit()

        if self._lcd is not None:
            try:
                self._lcd.parent.close()
            except Exception as e:
                self.logger.exception(str(e))

        try:
            if self.ui is not None:
                self.ui.close()

        except Exception as e:
            self.logger.exception(str(e))

        if __name__ == '__main__':
            self.parent.close()

    #  #####################################
    #  Methods for running the acquisition

    def init_hardware(self, do_init=True):
        """ Init the selected detector

        Parameters
        ----------
        do_init: bool
            If True, create a DAQ_Detector instance and move it into a separated thread, connected its signals/slots
            to the DAQ_Viewer object (self)
            If False, force the instrument to close and kill the Thread (still not done properly in some cases)
        """
        if not do_init:
            try:
                self.command_hardware.emit(ThreadCommand(command="close"))
                QtWidgets.QApplication.processEvents()
                if self.ui is not None:
                    self.ui.detector_init = False

            except Exception as e:
                self.logger.exception(str(e))
        else:            
            try:

                hardware = DAQ_Detector(self._title, self.settings, self.detector)
                self._hardware_thread = QThread()
                if self.config('viewer', 'viewer_in_thread'):
                    hardware.moveToThread(self._hardware_thread)

                self.command_hardware[ThreadCommand].connect(hardware.queue_command)
                hardware.data_detector_sig[DataToExport].connect(self.show_data)
                hardware.data_detector_temp_sig[DataToExport].connect(self.show_temp_data)
                hardware.status_sig[ThreadCommand].connect(self.thread_status)
                self._update_settings_signal[edict].connect(hardware.update_settings)

                self._hardware_thread.hardware = hardware
                if self.config('viewer', 'viewer_in_thread'):
                    self._hardware_thread.start()
                self.command_hardware.emit(ThreadCommand("ini_detector", attribute=[
                    self.settings.child('detector_settings').saveState(), self.controller]))
                if self.ui is not None:
                    for dock in self.ui.viewer_docks:
                        dock.setEnabled(True)

            except Exception as e:
                self.logger.exception(str(e))

    def snap(self):
        """ Launch a single grab """
        self.grab_data(False, snap_state=True)

    def grab(self):
        """ Launch a continuous grab """
        if self.ui is not None:
            self.manage_ui_actions('grab', 'setChecked', not self._grabing)
            self.grab_data(not self._grabing, snap_state=False)

    def snapshot(self, pathname=None, dosave=False, send_to_tcpip=False):
        """Do one single grab (snap) and eventually save the data.

        Parameters
        ----------
        pathname: str or Path object
            The path where to save data
        dosave: bool
            Do save or just grab data
        send_to_tcpip: bool
            If True, send the grabbed data through the TCP/IP pipe
        """
        try:
            self._do_save_data = dosave
            if pathname is None:
                raise (ValueError("filepathanme has not been defined in snapshot"))

            self._save_file_pathname = pathname
            self.grab_data(grab_state=False, send_to_tcpip=send_to_tcpip, snap_state=True)
        except Exception as e:
            self.logger.exception(str(e))

    def grab_data(self, grab_state=False, send_to_tcpip=False, snap_state=False):
        """ Generic method to grab or snap data from the selected (and initialized) detector

        Parameters
        ----------
        grab_state: bool
            Defines the grab status: if True: do live grabbing if False stops the grab
        send_to_tcpip: bool
            If True, send the grabbed data through the TCP/IP pipe
        snap_state: bool
            if True performs a single grab
        """
        self._grabing = grab_state
        self._send_to_tcpip = send_to_tcpip
        self._grab_done = False

        if self.ui is not None:
            self.ui.data_ready = False

        self._start_grab_time = time.perf_counter()
        if snap_state:
            self.update_status(f'{self._title}: Snap')
            self.command_hardware.emit(
                ThreadCommand("single", dict(Naverage=self.settings['main_settings', 'Naverage'])))
        else:
            if not grab_state:
                self.update_status(f'{self._title}: Stop Grab')
                self.command_hardware.emit(ThreadCommand("stop_grab", ))
            else:
                self.thread_status(ThreadCommand("update_channels", ))
                self.update_status(f'{self._title}: Continuous Grab')
                self.command_hardware.emit(
                    ThreadCommand("grab", dict(Naverage=self.settings['main_settings', 'Naverage'])))

    def take_bkg(self):
        """ Do a snap and store data to be used as background into an attribute: `self._bkg`

        The content of the bkg will be saved if data is further saved with do_bkg property set to True
        """
        self._take_bkg = True
        self.grab_data(snap_state=True)

    def stop_grab(self):
        """ Stop the current continuous grabbing and unchecked the stop button of the UI

        See Also
        --------
        :meth:`stop`
        """
        if self.ui is not None:
            self.manage_ui_actions('grab', 'setChecked', False)
        self.stop()

    def stop(self):
        """ Stop the current continuous grabbing """
        self.update_status(f'{self._title}: Stop Grab')
        self.command_hardware.emit(ThreadCommand("stop_all", ))
        self._grabing = False

    @Slot()
    def _raise_timeout(self):
        """  Print the "timeout occurred" error message in the status bar via the update_status method.
        """
        self.update_status("Timeout occured", log_type="log")

    @staticmethod
    def load_data():
        """Opens a H5 file in the H5Browser module

        Convenience static method.
        """
        browse_data()

    def save_current(self):
        """Save current data into a h5file"""
        self._do_save_data = True
        self._save_file_pathname = select_file(start_path=self._save_file_pathname, save=True,
                                                                                  ext='h5')  # see daq_utils
        self._save_export_data(self._data_to_save_export)

    def save_new(self):
        """Snap data and save them into a h5file"""
        self._do_save_data = True
        self._save_file_pathname = select_file(start_path=self._save_file_pathname, save=True,
                                                                                  ext='h5')  # see daq_utils
        self.snapshot(pathname=self._save_file_pathname, dosave=True)

    def _init_continuous_save(self):
        """ Initialize the continuous saving H5Saver object

        Update the module_and_data_saver attribute as :class:`DetectorEnlargeableSaver` object
        """
        if self._h5saver_continuous.settings.child('do_save').value():

            self._h5saver_continuous.settings.child('base_name').setValue('Data')
            self._h5saver_continuous.settings.child('N_saved').show()
            self._h5saver_continuous.settings.child('N_saved').setValue(0)
            self.module_and_data_saver.h5saver = self._h5saver_continuous
            self._h5saver_continuous.init_file(update_h5=True)

            self.module_and_data_saver = module_saving.DetectorEnlargeableSaver(self)
            self.module_and_data_saver.h5saver = self._h5saver_continuous
            self.module_and_data_saver.get_set_node()

            self.grab_done_signal.connect(self.append_data)
        else:
            self._do_continuous_save = False
            self._h5saver_continuous.settings.child('N_saved').hide()
            self.grab_done_signal.disconnect(self.append_data)

            try:
                self._h5saver_continuous.close()
            except Exception as e:
                self.logger.exception(str(e))

    def append_data(self, dte: DataToExport = None, where: Union[Node, str] = None):
        """Appends current DataToExport to a DetectorEnlargeableSaver

        Method to be used when performing continuous saving into a h5file (continuous mode or DAQ_Logger)

        Parameters
        ----------
        dte: DataToExport
            not really used
        where: Node or str
        See Also
        --------
        :class:`DetectorEnlargeableSaver`
        """
        if dte is None:
            dte = self._data_to_save_export
        self._add_data_to_saver(dte, init_step=self._h5saver_continuous.settings['N_saved'] == 0,
                                where=where)
        self._h5saver_continuous.settings.child('N_saved').setValue(self._h5saver_continuous.settings['N_saved'] + 1)

    def insert_data(self, indexes: Tuple[int], where: Union[Node, str] = None,
                    distribution=DataDistribution['uniform']):
        """Insert DataToExport to a DetectorExtendedSaver at specified indexes

        Method to be used when saving into an already initialized array within a h5file (DAQ_Scan for instance)

        Parameters
        ----------
        indexes: tuple(int)
            The indexes within the extended array where to place these data
        where: Node or str
        distribution: DataDistribution enum

        See Also
        --------
        DAQ_Scan, DetectorExtendedSaver
        """
        self._add_data_to_saver(self._data_to_save_export, init_step=np.all(np.array(indexes) == 0), where=where,
                                indexes=indexes, distribution=distribution)

    def _add_data_to_saver(self, dte: DataToExport, init_step=False, where=None, **kwargs):
        """Adds DataToExport data to the current node using the declared module_and_data_saver

        Filters the data to be saved by DataSource as specified in the current H5Saver (see self.module_and_data_saver)

        Parameters
        ----------
        dte: DataToExport
            The data to be saved
        init_step: bool
            If True, means this is the first step of saving (if multisaving), then save background if any and a png image
        kwargs: dict
            Other named parameters to be passed as is to the module_and_data_saver

        See Also
        --------
        DetectorSaver, DetectorEnlargeableSaver, DetectorExtendedSaver

        """
        detector_node = self.module_and_data_saver.get_set_node(where)
        dte = dte if not self.module_and_data_saver.h5saver.settings['save_raw_only'] else \
            dte.get_data_from_source('raw')  # filters depending on the source: raw or calculated

        dte = DataToExport(name=dte.name, data=  # filters depending on the extra argument 'save'
                           [dwa for dwa in dte if ('save' not in dwa.extra_attributes) or
                            ('save' in dwa.extra_attributes and dwa.save)])

        self.module_and_data_saver.add_data(detector_node, dte, **kwargs)

        if init_step:
            if self._do_bkg and self._bkg is not None:
                self.module_and_data_saver.add_bkg(detector_node, self._bkg)

    def _save_data(self, path=None, dte: DataToExport = None):
        """Private. Practical implementation to save data into a h5file altogether with metadata, axes, background...

        Parameters
        ----------
        path: Path
            where to save the data as returned from browse_file for instance
        dte: DataToExport

        See Also
        --------
        browse_file, _get_data_from_viewers
        """
        if path is not None:
            path = Path(path)
        h5saver = H5Saver(save_type='detector')
        h5saver.init_file(update_h5=True, custom_naming=False, addhoc_file_path=path)
        self.module_and_data_saver = module_saving.DetectorSaver(self)
        self.module_and_data_saver.h5saver = h5saver

        self._add_data_to_saver(dte, init_step=True)

        if self.ui is not None:
            (root, filename) = os.path.split(str(path))
            filename, ext = os.path.splitext(filename)
            image_path = os.path.join(root, filename + '.png')
            self.dockarea.parent().grab().save(image_path)

        h5saver.close_file()
        self.data_saved.emit()

    @Slot(DataToExport)
    def _save_export_data(self, data: DataToExport):
        """Auxiliary method (Slot) to receive all data (raw and processed from rois) and save them

        Parameters
        ----------
        data: DataToExport

        See Also
        --------
        _save_data
        """

        if self._do_save_data:
            self._save_data(self._save_file_pathname, data)
            self._do_save_data = False

    def _get_data_from_viewer(self, data: DataToExport):
        """Get all data emitted by the current viewers

        Each viewer *data_to_export_signal* is connected to this slot. The collected data is stored in another
        DataToExport `self._data_to_save_export` for further processing. All raw data are also stored in this attribute.
        When all viewers have emitted this signal, the collected data are emitted  with the
        `grab_done_signal` signal.

        Parameters
        ---------_
        data: DataToExport
            All data collected from the viewers

        """
        if self._data_to_save_export is not None:  # means that somehow data are not initialized so no further procsessing
            self._received_data += 1
            if len(data) != 0:
                for dat in data:
                    dat.origin = f'{self.title} - {dat.origin}' if dat.origin is not None else f'{self.title}'
                self._data_to_save_export.append(data)

            if self._received_data == len(self.viewers):
                self._grab_done = True
                self.grab_done_signal.emit(self._data_to_save_export)

    @property
    def current_data(self) -> DataToExport:
        """ Get the current data stored internally"""
        return self._data_to_save_export

    @Slot(DataToExport)
    def show_temp_data(self, data: DataToExport):
        """Send data to their dedicated viewers but those will not emit processed data signal

        Slot receiving data from plugins emitted with the `data_grabed_signal_temp`

        Parameters
        ----------
        data: list of DataFromPlugins
        """
        self._init_show_data(data)
        if self.ui is not None:
            self.set_data_to_viewers(data, temp=True)

    @Slot(DataToExport)
    def show_data(self, dte: DataToExport):
        """Send data to their dedicated viewers

        Slot receiving data from plugins emitted with the `data_grabed_signal`
        Process the data as specified in the settings, display them into the dedicated data viewers depending on the
        settings:
            * create a container (OrderedDict `_data_to_save_export`) with info from this DAQ_Viewer (title), a timestamp...
            * call `_process_data`
            * do background subtraction if any
            * check refresh time (if set in the settings) to send or not data to data viewers
            * either send to the data viewers (if refresh time is ok and/or show data option in settings is set)
            * either
                * send grab_done_signal (to the slot _save_export_data ) to save the data

        Parameters
        ----------
        dte: DataToExport

        See Also
        --------
        _init_show_data, _process_data
        """
        try:
            dte = dte.deepcopy()
            if self.settings.child('main_settings', 'tcpip', 'tcp_connected').value() and self._send_to_tcpip:
                self._command_tcpip.emit(ThreadCommand('data_ready', dte))
            if self.ui is not None:
                self.ui.data_ready = True

            if self.settings['main_settings', 'live_averaging']:
                self.settings.child('main_settings', 'N_live_averaging').setValue(self._ind_continuous_grab)
                _current_data = dte.deepcopy()

                self._ind_continuous_grab += 1
                if self._ind_continuous_grab > 1:
                    self._data_to_save_export = \
                        _current_data.average(self._data_to_save_export, self._ind_continuous_grab)
            else:
                for dwa in dte:
                    dwa.origin = self._title
                self._data_to_save_export = DataToExport(self._title, control_module='DAQ_Viewer', data=dte.data)

            if self._take_bkg:
                self._bkg = self._data_to_save_export.deepcopy()
                self._take_bkg = False

            if self._grabing:  # if live
                refresh_time = self.settings['main_settings', 'refresh_time']
                refresh = time.perf_counter() - self._start_grab_time > refresh_time / 1000
                if refresh:
                    self._start_grab_time = time.perf_counter()
            else:
                refresh = True  # if single
            if self.ui is not None and self.settings.child('main_settings', 'show_data').value() and refresh:
                self._received_data = 0  # so that data send back from viewers can be properly counted
                data_to_plot = self._data_to_save_export.get_data_from_attribute('plot', True, deepcopy=True)
                data_to_plot.append(self._data_to_save_export.get_data_from_missing_attribute('plot', deepcopy=True))
                # process bkg if needed
                if self.do_bkg and self._bkg is not None:
                    data_to_plot -= self._bkg

                self._init_show_data(data_to_plot)
                self.set_data_to_viewers(data_to_plot)
            else:
                self._grab_done = True
                self.grab_done_signal.emit(self._data_to_save_export)

        except Exception as e:
            self.logger.exception(str(e))

    def _init_show_data(self, dte: DataToExport):
        """Processing before showing data

        * process the data to check if they overshoot
        * check the data dimensionality to update the dedicated viewers

        Parameters
        ----------
        dte: DataToExport

        See Also
        --------
        _process_overshoot
        """
        self._process_overshoot(dte)

        self._viewer_types = [ViewersEnum(dwa.dim.name) for dwa in dte if
                              ('plot' not in dwa.extra_attributes) or
                              ('plot' in dwa.extra_attributes and dwa.plot)]
        if self.ui is not None:
            if self.ui.viewer_types != self._viewer_types:
                self.ui.update_viewers(self._viewer_types)

    def set_data_to_viewers(self, dte: DataToExport, temp=False):
        """Process data dimensionality and send appropriate data to their data viewers

        Parameters
        ----------
        dte: DataToExport
        temp: bool
            if True notify the data viewers to display data as temporary (meaning not exporting processed data from roi)

        See Also
        --------
        ViewerBase, Viewer0D, Viewer1D, Viewer2D
        """
        for ind, dwa in enumerate(dte):
            if ('plot' not in dwa.extra_attributes) or \
                    ('plot' in dwa.extra_attributes and dwa.plot):
                self.viewers[ind].title = dwa.name
                self.viewer_docks[ind].setTitle(self._title + ' ' + dwa.name)

                if temp:
                    self.viewers[ind].show_data_temp(dwa)
                else:
                    self.viewers[ind].show_data(dwa)

    def value_changed(self, param: Parameter):
        """ParameterManager subclassed method. Process events from value changed by user in the UI Settings

        Parameters
        ----------
        param: Parameter
            a given parameter whose value has been changed by user
        """
        path = self.settings.childPath(param)
        if param.name() == 'DAQ_type':
            self._h5saver_continuous.settings.child('do_save').setValue(False)
            self.settings.child('main_settings', 'axes').show(param.value() == 'DAQ2D')

        elif param.name() == 'show_averaging':
            self.settings.child('main_settings', 'live_averaging').setValue(False)
            self._update_settings_signal.emit(edict(path=path, param=param, change='value'))

        elif param.name() == 'live_averaging':
            self.settings.child('main_settings', 'show_averaging').setValue(False)
            if param.value():
                self.settings.child('main_settings', 'N_live_averaging').show()
                self._ind_continuous_grab = 0
                self.settings.child('main_settings', 'N_live_averaging').setValue(0)
            else:
                self.settings.child('main_settings', 'N_live_averaging').hide()
            #self._update_settings_signal.emit(edict(path=path, param=param, change='value'))

        elif param.name() in putils.iter_children(self.settings.child('main_settings', 'axes'), []):
            if self.daq_type.name == "DAQ2D":
                if param.name() == 'use_calib':
                    if param.value() != 'None':
                        params = ioxml.XML_file_to_parameter(
                            os.path.join(local_path, 'camera_calibrations', param.value() + '.xml'))
                        param_obj = Parameter.create(name='calib', type='group', children=params)
                        self.settings.child('main_settings', 'axes').restoreState(
                            param_obj.child('axes').saveState(), addChildren=False, removeChildren=False)
                        self.settings.child('main_settings', 'axes').show()
                else:
                    for viewer in self.viewers:
                        viewer.x_axis, viewer.y_axis = self.get_scaling_options()

        elif param.name() == 'continuous_saving_opt':
            self._h5saver_continuous.show_settings(param.value())

        elif param.name() == 'wait_time':
            self.command_hardware.emit(ThreadCommand('update_wait_time', [param.value()]))

        elif param.name() == 'connect_server':
            if param.value():
                self.connect_tcp_ip()
            else:
                self._command_tcpip.emit(ThreadCommand('quit', ))

        elif param.name() == 'ip_address' or param.name == 'port':
            self._command_tcpip.emit(
                ThreadCommand('update_connection',
                              dict(ipaddress=self.settings['main_settings', 'tcpip', 'ip_address'],
                                   port=self.settings['main_settings', 'tcpip', 'port'])))

        elif param.name() == 'plugin_config':
            self.show_config(self.plugin_config)

        if path is not None:
            if 'main_settings' not in path:
                self._update_settings_signal.emit(edict(path=path, param=param, change='value'))

                if self.settings.child('main_settings', 'tcpip', 'tcp_connected').value():
                    self._command_tcpip.emit(ThreadCommand('send_info', dict(path=path, param=param)))

    def child_added(self, param, data):
        """ Adds a child in the settings attribute

        Parameters
        ----------
        param: Parameter
            the parameter where child will be added
        data: Parameter
            the child parameter
        """
        if param.name() not in putils.iter_children(self.settings.child('main_settings'), []):
            self._update_settings_signal.emit(edict(path=putils.get_param_path(param)[1:], param=data[0],
                                                    change='childAdded'))

    def param_deleted(self, param):
        """ Remove a child from the settings attribute

        Parameters
        ----------
        param: Parameter
            a given parameter whose value has been changed by user
        """
        if param.name() not in putils.iter_children(self.settings.child('main_settings'), []):
            self._update_settings_signal.emit(edict(path=['detector_settings'], param=param, change='parent'))

    def _set_setting_tree(self):
        """Apply the specific settings of the selected detector (plugin)

        Remove previous ones and load on the fly the new ones

        See Also
        --------
        pymodaq.control_modules.utils:get_viewer_plugins
        """

        try:
            if len(self.settings.child('detector_settings').children()) > 0:
                for child in self.settings.child('detector_settings').children():
                    child.remove()

            det_params, _class = get_viewer_plugins(self.daq_type.name, self.detector)
            self.settings.child('detector_settings').addChildren(det_params.children())
        except Exception as e:
            self.logger.exception(str(e))

    def _process_overshoot(self, dte: DataToExport):
        """Compare data value (0D) to the given overshoot setting
        """
        if self.settings.child('main_settings', 'overshoot', 'stop_overshoot').value():
            for dwa in dte:
                for data_array in dwa.data:
                    if any(data_array >= self.settings.child('main_settings', 'overshoot', 'overshoot_value').value()):
                        self.overshoot_signal.emit(True)

    def get_scaling_options(self):
        """Create axes scaling options depending on the ('main_settings', 'axes') settings

        Returns
        -------
        Tuple[Axis]
        """
        scaled_xaxis = Axis(label=self.settings['main_settings', 'axes', 'xaxis', 'xlabel'],
                            units=self.settings['main_settings', 'axes', 'xaxis', 'xunits'],
                            offset=self.settings['main_settings', 'axes', 'xaxis', 'xoffset'],
                            scaling=self.settings['main_settings', 'axes', 'xaxis', 'xscaling'])
        scaled_yaxis = Axis(label=self.settings['main_settings', 'axes', 'yaxis', 'ylabel'],
                            units=self.settings['main_settings', 'axes', 'yaxis', 'yunits'],
                            offset=self.settings['main_settings', 'axes', 'yaxis', 'yoffset'],
                            scaling=self.settings['main_settings', 'axes', 'yaxis', 'yscaling'])
        return scaled_xaxis, scaled_yaxis

    def thread_status(self, status: ThreadCommand):
        """Get back info (using the ThreadCommand object) from the hardware

        And re-emit this ThreadCommand using the custom_sig signal if it should be used in a higher level module

        Commands valid for all control modules are defined in the parent class, here are described only the specific
        ones

        Parameters
        ----------
        status: ThreadCommand
            The info returned from the hardware, the command (str) can be either:
                * ini_detector: update the status with "detector initialized" value and init state if attribute not null.
                * grab : emit grab_status(True)
                * grab_stopped: emit grab_status(False)
                * init_lcd: display a LCD panel
                * lcd: display on the LCD panel, the content of the attribute
                * stop: stop the grab
        """
        super().thread_status(status, 'detector')

        if status.command == "ini_detector":
            self.update_status("detector initialized: " + str(status.attribute[0]['initialized']))
            if self.ui is not None:
                self.ui.detector_init = status.attribute[0]['initialized']
            if status.attribute[0]['initialized']:
                self.controller = status.attribute[0]['controller']
                self._initialized_state = True
            else:
                self._initialized_state = False

            self.init_signal.emit(self._initialized_state)

        elif status.command == "grab":
            self.grab_status.emit(True)

        elif status.command == 'grab_stopped':
            self.grab_status.emit(False)

        elif status.command == 'init_lcd':
            if self._lcd is not None:
                try:
                    self._lcd.parent.close()
                except Exception as e:
                    self.logger.exception(str(e))
            # lcd module
            lcd = QtWidgets.QWidget()
            self._lcd = LCD(lcd, **status.attribute)
            lcd.setVisible(True)
            QtWidgets.QApplication.processEvents()

        elif status.command == 'lcd':
            self._lcd.setvalues(status.attribute)

        elif status.command == 'stop':
            self.stop_grab()

    def connect_tcp_ip(self):
        """Init a TCPClient in a separated thread to communicate with a distant TCp/IP Server

        Use the settings: ip_address and port to specify the connection

        See Also
        --------
        TCPServer
        """
        if self.settings.child('main_settings', 'tcpip', 'connect_server').value():
            self._tcpclient_thread = QThread()

            tcpclient = TCPClient(self.settings.child('main_settings', 'tcpip', 'ip_address').value(),
                                  self.settings.child('main_settings', 'tcpip', 'port').value(),
                                  self.settings.child('detector_settings'))
            tcpclient.moveToThread(self._tcpclient_thread)
            self._tcpclient_thread.tcpclient = tcpclient
            tcpclient.cmd_signal.connect(self.process_tcpip_cmds)

            self._command_tcpip[ThreadCommand].connect(tcpclient.queue_command)

            self._tcpclient_thread.started.connect(tcpclient.init_connection)

            self._tcpclient_thread.start()

    @Slot(ThreadCommand)
    def process_tcpip_cmds(self, status):
        """Receive commands from the TCP Server (if connected) and process them

        Parameters
        ----------
        status: ThreadCommand
            Possible commands are:
            * 'Send Data: to trigger a snapshot
            * 'connected': show that connection is ok
            * 'disconnected': show that connection is not OK
            * 'Update_Status': update a status command
            * 'set_info': receive settings from the server side and update them on this side

        See Also
        --------
        connect_tcp_ip, TCPServer

        """
        if 'Send Data' in status.command:
            self.snapshot('', send_to_tcpip=True)
        elif status.command == 'connected':
            self.settings.child('main_settings', 'tcpip', 'tcp_connected').setValue(True)

        elif status.command == 'disconnected':
            self.settings.child('main_settings', 'tcpip', 'tcp_connected').setValue(False)

        elif status.command == 'Update_Status':
            self.thread_status(status)

        elif status.command == 'set_info':
            param_dict = ioxml.XML_string_to_parameter(status.attribute[1])[0]
            param_tmp = Parameter.create(**param_dict)
            param = self.settings.child('detector_settings', *status.attribute[0][1:])

            param.restoreState(param_tmp.saveState())

        elif status.command == 'get_axis':
            raise DeprecationWarning('Do not use this, the axis are in the data objects')
            self.command_hardware.emit(
                ThreadCommand('get_axis', ))  # tells the plugin to emit its axes so that the server will receive them


class DAQ_Detector(QObject):
    """ Worker class to control the instrument plugin

    Attributes
    ----------
    detector: real instance of the instrument plugin class
    controller: DAQ_Viewer_base
        wrapper object used to control a given instrument in the instrument plugin
    controller_adress: int
        unique integer used to identify a controller shared among multiple instrument plugins

    """
    status_sig = Signal(ThreadCommand)
    data_detector_sig = Signal(DataToExport)
    data_detector_temp_sig = Signal(DataToExport)

    def __init__(self, title, settings_parameter, detector_name):
        super().__init__()
        self.waiting_for_data = False
        self.controller = None
        self.logger = set_logger(f'{logger.name}.{title}.detector')
        self._title = title
        self.detector_name = detector_name
        self.detector: DAQ_Viewer_base = None
        self.controller_adress: int = None
        self.grab_state = False
        self.single_grab = False
        self.datas: DataToExport = None
        self.ind_average = 0
        self.Naverage = 1
        self.average_done = False
        self.hardware_averaging = False
        self.show_averaging = False
        self.wait_time = settings_parameter['main_settings', 'wait_time']
        self.daq_type = DAQTypesEnum[settings_parameter['main_settings', 'DAQ_type']]

    @property
    def title(self):
        return self._title

    def update_settings(self, settings_parameter_dict):
        """ Apply a Parameter serialized as a dict to the instrument plugin class or to self

        Parameters
        ----------
        settings_parameter_dict: dict
            dictionary serializing a Parameter object

        Examples
        --------
        If the parameter is of the form ('detector_settings', 'xxx') then the parameter is sent to the instrument
        plugin class.
        """

        path = settings_parameter_dict['path']
        param = settings_parameter_dict['param']
        if path[0] == 'main_settings':
            if hasattr(self, path[-1]):
                setattr(self, path[-1], param.value())

        elif path[0] == 'detector_settings':
            self.detector.update_settings(settings_parameter_dict)

    def queue_command(self, command: ThreadCommand):
        """Transfer command from the main module to the hardware module

        Parameters
        ----------
        command: ThreadCommand
            The specific (or generic) command (str) to pass to the hardware,  either:
            * ini_detector
            * close
            * grab
            * single
            * stop_grab
            * stop_all
            * update_scanner
            * move_at_navigator
            * update_wait_time
            * get_axis
            * any string that the hardware is able to understand
        """
        if command.command == "ini_detector":
            status = self.ini_detector(*command.attribute)
            self.status_sig.emit(ThreadCommand(command.command, [status, 'log']))

        elif command.command == "close":
            status = self.close()
            self.status_sig.emit(ThreadCommand(command.command, [status, 'log']))

        elif command.command == "grab":
            self.single_grab = False
            self.grab_state = True
            self.grab_data(**command.attribute)

        elif command.command == "single":
            self.single_grab = True
            self.grab_state = True
            self.single(**command.attribute)

        elif command.command == "stop_grab":
            self.grab_state = False
            self.status_sig.emit(ThreadCommand("Update_Status", ['Stoping grab']))

        elif command.command == "stop_all":
            self.grab_state = False
            self.detector.stop()
            QtWidgets.QApplication.processEvents()
            self.status_sig.emit(ThreadCommand("Update_Status", ['Stoping grab']))

        elif command.command == 'update_scanner':
            self.detector.update_scanner(command.attribute[0])

        elif command.command == 'move_at_navigator':
            self.detector.move_at_navigator(*command.attribute)

        elif command.command == 'update_wait_time':
            self.wait_time = command.attribute[0]

        elif command.command == 'get_axis':
            self.detector.get_axis()

        else:  # custom commands for particular plugins
            if hasattr(self.detector, command.command):
                cmd = getattr(self.detector, command.command)
                cmd(command.attribute)

    def ini_detector(self, params_state=None, controller=None):
        """ Initialize an instrument plugin class and tries to apply preset settings

        When the instrument is initialized from the Dashboard using a Preset, tries to apply the preset
        settings to the instrument instance

        Parameters
        ----------
        params_state: dict
        controller: wrapper
        """
        try:
            # status="Not initialized"
            status = edict(initialized=False, info="", x_axis=None, y_axis=None)
            det_params, class_ = get_viewer_plugins(self.daq_type.name, self.detector_name)
            self.detector: DAQ_Viewer_base = class_(self, params_state)

            try:
                self.detector.dte_signal.connect(self.data_ready)
                self.detector.dte_signal_temp.connect(self.emit_temp_data)
                infos = self.detector.ini_detector(controller)
                status.controller = self.detector.controller

            except Exception as e:
                logger.exception('Hardware couldn\'t be initialized' + str(e))
                infos = str(e), False
                status.controller = None

            if isinstance(infos, edict):
                status.update(infos)
            else:
                status.info = infos[0]
                status.initialized = infos[1]

            self.hardware_averaging = class_.hardware_averaging  # to check if averaging can be done directly by
            # the hardware or done here software wise

            return status
        except Exception as e:
            self.logger.exception(str(e))
            return status

    def emit_temp_data(self, data: DataToExport):
        """ Convenience method to export temporary data using the data_detector_temp_sig Signal

        Parameters
        ----------
        data: DataToExport
        """
        self.data_detector_temp_sig.emit(data)

    def data_ready(self, data: DataToExport):
        """ Process the data received from the instrument plugin class

        Processing here is eventual software averaging if it was not possible in the instrument plugin class

        Parameters
        ----------
        data: DataToExport
        """
        do_averaging = self.Naverage > 1 and not self.hardware_averaging

        if do_averaging:  # to execute if the averaging has to be done software wise
            self.ind_average += 1
            if self.ind_average == 1:
                self.datas = data.deepcopy()
            else:
                self.datas = data.average(self.datas, self.ind_average)

            if self.show_averaging:
                self.emit_temp_data(self.datas)

            if self.ind_average == self.Naverage:
                self.average_done = True
                self.data_detector_sig.emit(self.datas)
                self.ind_average = 0
        else:
            self.average_done = True  # expected to make sure the single_grab stop by itself
            self.data_detector_sig.emit(data)
        self.waiting_for_data = False
        if not self.grab_state:
            self.detector.stop()

    def single(self, Naverage=1, *args, **kwargs):
        """ Convenience function to grab a single set of data

        Parameters
        ----------
        Naverage: int
            The number of data to average before displaying
        kwargs: optional named arguments
        """
        self.grab_data(Naverage, live=False, **kwargs)

    def grab_data(self, Naverage=1, live=True, **kwargs):
        """ General method to grab data from the instrument plugin class

        Will check if the plugin class can do hardware averaging (if NAverage > 1) and and live_mode, otherwise
        do both software wise here

        Parameters
        ----------
        Naverage: int
            The number of data to average
        live: bool
            Try to run the instrument plugin class grabbing in live mode
        kwargs: optional named arguments passed to the grab_data method of the instrument plugin class
        """
        try:
            self.ind_average = 0
            self.Naverage = Naverage
            if Naverage > 1:
                self.average_done = False
            self.waiting_for_data = False

            # for live mode:two possibilities: either snap one data and regrab softwarewise (while True) or if
            # self.detector.live_mode_available is True all data is continuously emitted from the plugin
            if self.detector.live_mode_available:
                kwargs['wait_time'] = self.wait_time
            else:
                kwargs['wait_time'] = 0
            self.status_sig.emit(ThreadCommand('grab'))
            while True:
                try:
                    if not self.waiting_for_data:
                        self.waiting_for_data = True
                        self.detector.grab_data(Naverage, live=live, **kwargs)
                    QtWidgets.QApplication.processEvents()
                    if self.single_grab:
                        if self.hardware_averaging:
                            break
                        else:
                            if self.average_done:
                                break
                    else:
                        QThread.msleep(self.wait_time)  # if in grab mode apply a waiting time after acquisition
                    if not self.grab_state:
                        break   # if not in grab mode  breaks the while loop
                    if self.detector.live_mode_available and (not self.hardware_averaging and self.average_done):
                        break  # if live can be done in the plugin breaks the while loop except if average is asked but
                        # not done hardware wise
                except Exception as e:
                    self.logger.exception(str(e))
            self.status_sig.emit(ThreadCommand('grab_stopped'))

        except Exception as e:
            self.logger.exception(str(e))

    def close(self):
        """ Call the close method of the instrument plugin class
        """
        if self.detector is not None:
            status = self.detector.close()
            return status


def prepare_docks(area, title):
    """ Static method to init docks to be used within a DAQ_Viewer

    Parameters
    ----------
    area
    title

    Returns
    -------

    """
    dock_settings = Dock(title + " settings", size=(150, 250))
    dock_viewer = Dock(title + " viewer", size=(350, 350))
    area.addDock(dock_settings)
    area.addDock(dock_viewer, 'right', dock_settings)
    return dict(dock_settings=dock_settings, dock_viewer=dock_viewer)


def main(init_qt=True, init_det=False):
    """ Method called to start the DAQ_Viewer in standalone mode"""

    if init_qt:  # used for the test suite
        app = QtWidgets.QApplication(sys.argv)
        if config('style', 'darkstyle'):
            import qdarkstyle
            app.setStyleSheet(qdarkstyle.load_stylesheet(qdarkstyle.DarkPalette))

    win = QtWidgets.QMainWindow()
    area = DockArea()
    win.setCentralWidget(area)
    win.resize(1000, 500)
    win.setWindowTitle('PyMoDAQ Viewer')
    win.show()

    title = "Testing"
    viewer = DAQ_Viewer(area, title="Testing", daq_type=config('viewer', 'daq_type'),
                        **prepare_docks(area, title))
    if init_det:
        viewer.init_hardware_ui(init_det)

    if init_qt:
        sys.exit(app.exec_())
    return viewer, win


if __name__ == '__main__':
    main(init_det=False)



from qtpy import QtWidgets, QtCore
from pyqtgraph.widgets.SpinBox import SpinBox
from pyqtgraph.parametertree.parameterTypes.basetypes import WidgetParameterItem
from pymodaq.utils.parameter.utils import scroll_log, scroll_linear
import numpy as np


class SliderSpinBox(QtWidgets.QWidget):

    def __init__(self, *args, subtype='lin', **kwargs):

        super().__init__()
        self.subtype = subtype
        self.initUI(*args, **kwargs)

        self.valueChanged = self.spinbox.valueChanged  # (value)  for compatibility with QSpinBox
        self.sigValueChanged = self.spinbox.sigValueChanged  # (self)
        self.sigValueChanging = self.spinbox.sigValueChanging  # (self, value)  sent immediately; no delay.
        self.sigChanged = self.spinbox.sigValueChanged

    @property
    def opts(self):
        return self.spinbox.opts

    @opts.setter
    def opts(self, **opts):
        self.setOpts(**opts)

    def setOpts(self, **opts):
        self.spinbox.setOpts(**opts)
        if 'visible' in opts:
            self.slider.setVisible(opts['visible'])

    def insert_widget(self ,widget, row=0):
        self.vlayout.insertWidget(row, widget)

    def initUI(self, *args, **kwargs):
        """
            Init the User Interface.
        """
        self.vlayout = QtWidgets.QVBoxLayout()
        self.slider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        self.slider.setMinimumWidth(50)
        self.slider.setMinimum(0)
        self.slider.setMaximum(100)
        if 'value' in kwargs:
            value = kwargs.pop('value')
        else:
            if 'bounds' in kwargs:
                value = kwargs['bounds'][0]
            else:
                value = 1
        self.spinbox = SpinBox(parent=None, value=value, **kwargs)

        self.vlayout.addWidget(self.slider)
        self.vlayout.addWidget(self.spinbox)
        self.vlayout.setSpacing(0)
        self.vlayout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.vlayout)

        self.slider.valueChanged.connect(self.update_spinbox)
        self.spinbox.valueChanged.connect(self.update_slide)

    def update_spinbox(self, val):
        """
        val is a percentage [0-100] used in order to set the spinbox value between its min and max
        """
        min_val = float(self.opts['bounds'][0])
        max_val = float(self.opts['bounds'][1])
        if self.subtype == 'log':
            val_out = scroll_log(val, min_val, max_val)
        else:
            val_out = scroll_linear(val, min_val, max_val)
        try:
            self.slider.valueChanged.disconnect(self.update_spinbox)
            self.spinbox.valueChanged.disconnect(self.update_slide)
        except Exception:
            pass
        self.spinbox.setValue(val_out)

        self.slider.valueChanged.connect(self.update_spinbox)
        self.spinbox.valueChanged.connect(self.update_slide)

    def update_slide(self, val):
        """
        val is the spinbox value between its min and max
        """
        min_val = float(self.opts['bounds'][0])
        max_val = float(self.opts['bounds'][1])

        try:
            self.slider.valueChanged.disconnect(self.update_spinbox)
            self.spinbox.valueChanged.disconnect(self.update_slide)
        except Exception:
            pass
        if self.subtype == 'linear':
            value = int((val - min_val) / (max_val - min_val) * 100)
        else:
            value = int((np.log10(val) - np.log10(min_val)) / (np.log10(max_val) - np.log10(min_val)) * 100)
        self.slider.setValue(value)
        self.slider.valueChanged.connect(self.update_spinbox)
        self.spinbox.valueChanged.connect(self.update_slide)

    def setValue(self, val):
        self.spinbox.setValue(val)

    def value(self):
        return self.spinbox.value()


class SliderParameterItem(WidgetParameterItem):
    """Registered parameter type which displays a QLineEdit"""

    def makeWidget(self):
        opts = self.param.opts
        defs = {
            'value': 0, 'min': None, 'max': None,
            'step': 1.0, 'dec': False,
            'siPrefix': False, 'suffix': '', 'decimals': 12,
            'int': False
        }
        if 'subtype' not in opts:
            opts['subtype'] = 'linear'
        defs['bounds'] = [0., self.param.value()]  # max value set to default value when no max given
        if 'limits' not in opts:
            if 'min' in opts:
                defs['bounds'][0] = opts['min']
            if 'max' in opts:
                defs['bounds'][1] = opts['max']
        else:
            defs['bounds'] = opts['limits']

        if 'int' in opts:
            defs['int'] = opts['int']

        w = SliderSpinBox(subtype=opts['subtype'], bounds=defs['bounds'], value=defs['value'], int=defs['int'])
        self.setSizeHint(1, QtCore.QSize(50, 50))
        return w


from typing import Union
from qtpy import QtWidgets
from qtpy.QtCore import QObject, Slot, Signal

from pymodaq.utils.parameter import ioxml
from pymodaq.utils.parameter.utils import get_param_path, get_param_from_name, iter_children
from pymodaq.utils.parameter import Parameter
from easydict import EasyDict as edict

import numpy as np
from pymodaq.utils.math_utils import gauss1D, gauss2D
from pymodaq.utils.daq_utils import ThreadCommand, getLineInfo
from pymodaq.utils.config import Config, get_set_local_dir
from pymodaq.utils.tcp_ip.tcp_server_client import TCPServer, tcp_parameters
from pymodaq.utils.data import DataToExport, DataRaw
from pymodaq.utils.messenger import deprecation_msg
from pymodaq.utils.tcp_ip.mysocket import Socket
from pymodaq.utils.tcp_ip.serializer import DeSerializer, Serializer

comon_parameters = [{'title': 'Controller Status:', 'name': 'controller_status', 'type': 'list', 'value': 'Master',
                     'limits': ['Master', 'Slave']}, ]

local_path = get_set_local_dir()
# look for eventual calibration files
calibs = ['None']
if local_path.joinpath('camera_calibrations').is_dir():
    for ind_file, file in enumerate(local_path.joinpath('camera_calibrations').iterdir()):
        if 'xml' in file.suffix:
            calibs.append(file.stem)


config = Config()

params = [
    {'title': 'Main Settings:', 'name': 'main_settings', 'expanded': False, 'type': 'group', 'children': [
        {'title': 'DAQ type:', 'name': 'DAQ_type', 'type': 'list', 'limits': ['DAQ0D', 'DAQ1D', 'DAQ2D', 'DAQND'],
         'readonly': True},
        {'title': 'Detector type:', 'name': 'detector_type', 'type': 'str', 'value': '', 'readonly': True},
        {'title': 'Detector Name:', 'name': 'module_name', 'type': 'str', 'value': '', 'readonly': True},
        {'title': 'Plugin Config:', 'name': 'plugin_config', 'type': 'bool_push', 'label': 'Show Config', },
        {'title': 'Controller ID:', 'name': 'controller_ID', 'type': 'int', 'value': 0, 'default': 0, 'readonly': False},
        {'title': 'Show data and process:', 'name': 'show_data', 'type': 'bool', 'value': True, },
        {'title': 'Refresh time (ms):', 'name': 'refresh_time', 'type': 'float', 'value': 50., 'min': 0.},
        {'title': 'Naverage', 'name': 'Naverage', 'type': 'int', 'default': 1, 'value': 1, 'min': 1},
        {'title': 'Show averaging:', 'name': 'show_averaging', 'type': 'bool', 'default': False, 'value': False},
        {'title': 'Live averaging:', 'name': 'live_averaging', 'type': 'bool', 'default': False, 'value': False},
        {'title': 'N Live aver.:', 'name': 'N_live_averaging', 'type': 'int', 'default': 0, 'value': 0,
         'visible': False},
        {'title': 'Wait time (ms):', 'name': 'wait_time', 'type': 'int', 'default': 0, 'value': 00, 'min': 0},
        {'title': 'Continuous saving:', 'name': 'continuous_saving_opt', 'type': 'bool', 'default': False,
         'value': False},
        {'title': 'TCP/IP options:', 'name': 'tcpip', 'type': 'group', 'visible': True, 'expanded': False, 'children': [
            {'title': 'Connect to server:', 'name': 'connect_server', 'type': 'bool_push', 'label': 'Connect',
             'value': False},
            {'title': 'Connected?:', 'name': 'tcp_connected', 'type': 'led', 'value': False},
            {'title': 'IP address:', 'name': 'ip_address', 'type': 'str',
             'value': config('network', 'tcp-server', 'ip')},
            {'title': 'Port:', 'name': 'port', 'type': 'int', 'value': config('network', 'tcp-server', 'port')},
        ]},
        {'title': 'Overshoot options:', 'name': 'overshoot', 'type': 'group', 'visible': True, 'expanded': False,
         'children': [
             {'title': 'Overshoot:', 'name': 'stop_overshoot', 'type': 'bool', 'value': False},
             {'title': 'Overshoot value:', 'name': 'overshoot_value', 'type': 'float', 'value': 0}]},
        {'title': 'Axis options:', 'name': 'axes', 'type': 'group', 'visible': False, 'expanded': False, 'children': [
            {'title': 'Use calibration?:', 'name': 'use_calib', 'type': 'list', 'limits': calibs},
            {'title': 'X axis:', 'name': 'xaxis', 'type': 'group', 'children': [
                {'title': 'Label:', 'name': 'xlabel', 'type': 'str', 'value': "x axis"},
                {'title': 'Units:', 'name': 'xunits', 'type': 'str', 'value': "pxls"},
                {'title': 'Offset:', 'name': 'xoffset', 'type': 'float', 'default': 0., 'value': 0.},
                {'title': 'Scaling', 'name': 'xscaling', 'type': 'float', 'default': 1., 'value': 1.},
            ]},
            {'title': 'Y axis:', 'name': 'yaxis', 'type': 'group', 'children': [
                {'title': 'Label:', 'name': 'ylabel', 'type': 'str', 'value': "y axis"},
                {'title': 'Units:', 'name': 'yunits', 'type': 'str', 'value': "pxls"},
                {'title': 'Offset:', 'name': 'yoffset', 'type': 'float', 'default': 0., 'value': 0.},
                {'title': 'Scaling', 'name': 'yscaling', 'type': 'float', 'default': 1., 'value': 1.},
            ]},
        ]},

    ]},
    {'title': 'Detector Settings', 'name': 'detector_settings', 'type': 'group', 'children': []}
]


def main(plugin_file=None, init=True):
    """
    this method start a DAQ_Viewer object with this defined plugin as detector
    Returns
    -------
    """
    import sys
    from qtpy import QtWidgets
    from pymodaq.utils.gui_utils import DockArea
    from pymodaq.control_modules.daq_viewer import DAQ_Viewer
    from pathlib import Path

    app = QtWidgets.QApplication(sys.argv)
    if config('style', 'darkstyle'):
        import qdarkstyle
        app.setStyleSheet(qdarkstyle.load_stylesheet())

    win = QtWidgets.QMainWindow()
    area = DockArea()
    win.setCentralWidget(area)
    win.resize(1000, 500)
    win.setWindowTitle('PyMoDAQ Viewer')
    if plugin_file is None:
        detector = 'Mock'
        det_type = f'DAQ0D'
    else:
        detector = Path(plugin_file).stem[13:]
        det_type = f'DAQ{Path(plugin_file).stem[4:6].upper()}'
    prog = DAQ_Viewer(area, title="Testing")
    win.show()
    prog.daq_type = det_type
    prog.detector = detector
    if init:
        prog.init_hardware_ui()

    sys.exit(app.exec_())


class DAQ_Viewer_base(QObject):
    """
        ===================== ===================================
        **Attributes**          **Type**
        *hardware_averaging*    boolean
        *data_grabed_signal*    instance of Signal
        *params*                list
        *settings*              instance of pyqtgraph Parameter
        *parent*                ???
        *status*                dictionnary
        ===================== ===================================

        See Also
        --------
        send_param_status
    """
    hardware_averaging = False
    live_mode_available = False
    data_grabed_signal = Signal(list)  # will be deprecated use dte_signal
    data_grabed_signal_temp = Signal(list)  # will be deprecated use dte_signal_temp
    dte_signal = Signal(DataToExport)
    dte_signal_temp = Signal(DataToExport)

    params = []

    def __init__(self, parent=None, params_state=None):
        QObject.__init__(self)  # to make sure this is the parent class

        self.parent_parameters_path = []  # this is to be added in the send_param_status to take into account when
        # the current class instance parameter list is a child of some other class
        self.settings = Parameter.create(name='Settings', type='group', children=self.params)
        if params_state is not None:
            if isinstance(params_state, dict):
                self.settings.restoreState(params_state)
            elif isinstance(params_state, Parameter):
                self.settings.restoreState(params_state.saveState())

        if '0D' in str(self.__class__):
            self.plugin_type = '0D'
        elif '1D' in str(self.__class__):
            self.plugin_type = '1D'
        else:
            self.plugin_type = '2D'

        self.settings.sigTreeStateChanged.connect(self.send_param_status)

        self.parent = parent
        self.status = edict(info="", controller=None, initialized=False)
        self.scan_parameters = None

        self.x_axis = None
        self.y_axis = None

        self.controller = None

        if parent is not None:
            self._title = parent.title
        else:
            self._title = "mydetector"

        self.ini_attributes()

        self.data_grabed_signal.connect(self._emit_dte)
        self.data_grabed_signal_temp.connect(self._emit_dte_temp)

    def _emit_dte(self, dte: Union[DataToExport, list]):
        if isinstance(dte, list):
            deprecation_msg(f'Data emitted from the instrument plugins should be a DataToExport instance'
                            f'See: http://pymodaq.cnrs.fr/en/latest/developer_folder/'
                            f'instrument_plugins.html#emission-of-data')
            dte = DataToExport('temp', dte)
        self.dte_signal.emit(dte)

    def _emit_dte_temp(self, dte: Union[DataToExport, list]):
        if isinstance(dte, list):
            deprecation_msg(f'Data emitted from the instrument plugins should be a DataToExport instance'
                            f'See: http://pymodaq.cnrs.fr/en/latest/developer_folder/'
                            f'instrument_plugins.html#emission-of-data')
            dte = DataToExport('temp', dte)
        self.dte_signal_temp.emit(dte)

    def ini_attributes(self):
        """
        To be reimplemented in subclass
        """
        pass

    def ini_detector_init(self, old_controller=None, new_controller=None):
        """Manage the Master/Slave controller issue

        First initialize the status dictionnary
        Then check whether this stage is controlled by a multiaxe controller (to be defined for each plugin)
            if it is a multiaxes controller then:
            * if it is Master: init the controller here
            * if it is Slave: use an already initialized controller (defined in the preset of the dashboard)

        Parameters
        ----------
        old_controller: object
            The particular object that allow the communication with the hardware, in general a python wrapper around the
            hardware library. In case of Slave this one comes from a previously initialized plugin
        new_controller: object
            The particular object that allow the communication with the hardware, in general a python wrapper around the
            hardware library. In case of Master it is the new instance of your plugin controller
        """
        self.status.update(edict(info="", controller=None, initialized=False))
        if self.settings['controller_status'] == "Slave":
            if old_controller is None:
                raise Exception('no controller has been defined externally while this axe is a slave one')
            else:
                controller = old_controller
        else:  # Master stage
            controller = new_controller
        self.controller = controller
        return controller

    def ini_detector(self, controller=None):
        """
        Mandatory
        To be reimplemented in subclass
        """
        raise NotImplemented

    def close(self):
        """
        Mandatory
        To be reimplemented in subclass
        """
        raise NotImplemented

    def grab_data(self, Naverage=1, **kwargs):
        """
        Mandatory
        To be reimplemented in subclass
        """
        raise NotImplemented

    def stop(self):
        """
        Mandatory
        To be reimplemented in subclass
        """
        raise NotImplemented

    def commit_settings(self, param):
        """
        To be reimplemented in subclass
        """
        pass

    def get_axis(self):
        """deprecated"""
        raise DeprecationWarning('get_axis is deprecated add the axes within the DataWithAxes, '
                                 'DataFromPlugins')
        if self.plugin_type == '1D' or self.plugin_type == '2D':
            self.emit_x_axis()

        if self.plugin_type == '2D':
            self.emit_y_axis()

    def emit_status(self, status: ThreadCommand):
        """
            Emit the status signal from the given status.

            =============== ============ =====================================
            **Parameters**    **Type**     **Description**
            *status*                       the status information to transmit
            =============== ============ =====================================
        """
        if self.parent is not None:
            self.parent.status_sig.emit(status)
            QtWidgets.QApplication.processEvents()
        else:
            print(status)

    def update_scanner(self, scan_parameters):
        #todo check this because ScanParameters has been removed
        self.scan_parameters = scan_parameters

    @Slot(edict)
    def update_settings(self, settings_parameter_dict):
        """
            Update the settings tree from settings_parameter_dict.
            Finally do a commit to activate changes.

            ========================== ============= =====================================================
            **Parameters**              **Type**      **Description**
            *settings_parameter_dict*   dictionnnary  a dictionnary listing path and associated parameter
            ========================== ============= =====================================================

            See Also
            --------
            send_param_status, commit_settings
        """
        # settings_parameter_dict=edict(path=path,param=param)
        try:
            path = settings_parameter_dict['path']
            param = settings_parameter_dict['param']
            change = settings_parameter_dict['change']
            try:
                self.settings.sigTreeStateChanged.disconnect(self.send_param_status)
            except Exception:
                pass
            if change == 'value':
                self.settings.child(*path[1:]).setValue(param.value())  # blocks signal back to main UI
            elif change == 'childAdded':
                child = Parameter.create(name='tmp')
                child.restoreState(param.saveState())
                self.settings.child(*path[1:]).addChild(child)  # blocks signal back to main UI
                param = child

            elif change == 'parent':
                children = get_param_from_name(self.settings, param.name())

                if children is not None:
                    path = get_param_path(children)
                    self.settings.child(*path[1:-1]).removeChild(children)

            self.settings.sigTreeStateChanged.connect(self.send_param_status)

            self.commit_settings(param)
        except Exception as e:
            self.emit_status(ThreadCommand("Update_Status", [str(e), 'log']))



    def send_param_status(self, param, changes):
        """
            Check for changes in the given (parameter,change,information) tuple list.
            In case of value changed, send the 'update_settings' ThreadCommand with concerned path,data and change as attribute.

            =============== ============================================ ============================
            **Parameters**    **Type**                                    **Description**
            *param*           instance of pyqtgraph parameter             The parameter to check
            *changes*         (parameter,change,information) tuple list   The changes list to course
            =============== ============================================ ============================

            See Also
            --------
            daq_utils.ThreadCommand
        """
        for param, change, data in changes:
            path = self.settings.childPath(param)
            if change == 'childAdded':
                # first create a "copy" of the actual parameter and send this "copy", to be restored in the main UI
                self.emit_status(ThreadCommand('update_settings',
                                               [self.parent_parameters_path + path, [data[0].saveState(), data[1]],
                                                change]))  # send parameters values/limits back to the GUI. Send kind of a copy back the GUI otherwise the child reference will be the same in both th eUI and the plugin so one of them will be removed

            elif change == 'value' or change == 'limits' or change == 'options':
                self.emit_status(ThreadCommand('update_settings', [self.parent_parameters_path + path, data,
                                                                   change]))  # send parameters values/limits back to the GUI
            elif change == 'parent':
                pass

            pass



class DAQ_Viewer_TCP_server(DAQ_Viewer_base, TCPServer):
    """
        ================= ==============================
        **Attributes**      **Type**
        *command_server*    instance of Signal
        *x_axis*            1D numpy array
        *y_axis*            1D numpy array
        *data*              double precision float array
        ================= ==============================

        See Also
        --------
        utility_classes.DAQ_TCP_server
    """
    params_GRABBER = []  # parameters of a client grabber
    command_server = Signal(list)

    message_list = ["Quit", "Send Data 0D", "Send Data 1D", "Send Data 2D", "Send Data ND", "Status", "Done",
                    "Server Closed", "Info",
                    "Infos",
                    "Info_xml", 'x_axis', 'y_axis']
    socket_types = ["GRABBER"]
    params = comon_parameters + tcp_parameters

    def __init__(self, parent=None, params_state=None, grabber_type='2D'):
        """

        Parameters
        ----------
        parent
        params_state
        grabber_type: (str) either '0D', '1D' or '2D'
        """
        self.client_type = "GRABBER"
        DAQ_Viewer_base.__init__(self, parent, params_state)  # initialize base class with commom attribute and methods
        TCPServer.__init__(self, self.client_type)

        self.x_axis = None
        self.y_axis = None
        self.data = None
        self.grabber_type = grabber_type
        self.ind_data = 0
        self.data_mock: DataToExport = None

    def command_to_from_client(self, command: str):
        """Process the command"""
        sock = self.find_socket_within_connected_clients(self.client_type)
        if sock is not None:  # if client self.client_type is connected then send it the command

            if command == 'x_axis':
                raise DeprecationWarning(f'The command {command} is deprecated use the data objects')
            elif command == 'y_axis':
                raise DeprecationWarning(f'The command {command} is deprecated use the data objects')

            else:
                self.send_command(sock, command)

        else:  # else simulate mock data
            if command == "Send Data 0D":
                self.set_0D_Mock_data()
            elif command == "Send Data 1D":
                self.set_1D_Mock_data()
            elif command == "Send Data 2D":
                self.set_2D_Mock_data()
            self.process_cmds('Done')

    def send_data(self, sock: Socket, data: DataToExport):
        """
            To match digital and labview, send again a command.

            =============== ============================== ====================
            **Parameters**   **Type**                       **Description**
            *sock*                                          the socket receipt
            *data*           double precision float array   the data to be sent
            =============== ============================== ====================

            See Also
            --------
            send_command, check_send_data
        """
        self.send_command(sock, 'Done')
        sock.check_sended_with_serializer(data)

    def read_data(self, sock: Socket) -> DataToExport:
        """Read data from the socket
        """
        return DeSerializer(sock).dte_deserialization()

    def data_ready(self, data: DataToExport):
        """
            Send the grabed data signal.
        """
        self.dte_signal.emit(data)

    def command_done(self, command_sock):
        try:
            sock = self.find_socket_within_connected_clients(self.client_type)
            if sock is not None:  # if client self.client_type is connected then send it the command
                data: DataToExport = self.read_data(sock)
            else:
                data = self.data_mock

            if command_sock is None:
                self.data_ready(data)
            else:
                self.send_data(command_sock, data)  # to be send to a client

        except Exception as e:
            self.emit_status(ThreadCommand("Update_Status", [str(e), 'log']))

    def commit_settings(self, param):

        if param.name() in iter_children(self.settings.child(('settings_client')), []):
            grabber_socket: Socket = \
                [client['socket'] for client in self.connected_clients if client['type'] == self.client_type][0]
            grabber_socket.check_sended_with_serializer('set_info')

            path = get_param_path(param)[2:]  # get the path of this param as a list starting at parent 'infos'
            grabber_socket.check_sended_with_serializer(path)

            # send value
            data = ioxml.parameter_to_xml_string(param)
            grabber_socket.check_sended_with_serializer(data)

    def ini_detector(self, controller=None):
        """
            | Initialisation procedure of the detector updating the status dictionnary.
            |
            | Init axes from image , here returns only None values (to tricky to di it with the server and not really
             necessary for images anyway)

            See Also
            --------
            utility_classes.DAQ_TCP_server.init_server, get_xaxis, get_yaxis
        """
        self.settings.child('infos').addChildren(self.params_GRABBER)

        self.init_server()
        self.controller = self.serversocket
        # %%%%%%% init axes from image , here returns only None values (to tricky to di it with the server and not really necessary for images anyway)
        self.x_axis = self.get_xaxis()
        self.y_axis = self.get_yaxis()

        initialized = True
        info = "Server ready"
        return info, initialized

    def close(self):
        """
            Should be used to uninitialize hardware.

            See Also
            --------
            utility_classes.DAQ_TCP_server.close_server
        """
        self.listening = False
        self.close_server()

    def get_xaxis(self):
        """
            Obtain the horizontal axis of the image.

            Returns
            -------
            1D numpy array
                Contains a vector of integer corresponding to the horizontal camera pixels.
        """
        pass
        return self.x_axis

    def get_yaxis(self):
        """
            Obtain the vertical axis of the image.

            Returns
            -------
            1D numpy array
                Contains a vector of integer corresponding to the vertical camera pixels.
        """
        pass
        return self.y_axis

    def grab_data(self, Naverage=1, **kwargs):
        """
            Start new acquisition.
            Grabbed indice is used to keep track of the current image in the average.

            ============== ========== ==============================
            **Parameters**   **Type**  **Description**

            *Naverage*        int       Number of images to average
            ============== ========== ==============================

            See Also
            --------
            utility_classes.DAQ_TCP_server.process_cmds
        """
        try:
            self.ind_grabbed = 0  # to keep track of the current image in the average
            self.Naverage = Naverage
            self.process_cmds("Send Data {:s}".format(self.grabber_type))
            # self.command_server.emit(["process_cmds","Send Data 2D"])

        except Exception as e:
            self.emit_status(ThreadCommand('Update_Status', [getLineInfo() + str(e), "log"]))

    def stop(self):
        """
            not implemented.
        """
        pass
        return ""

    def set_0D_Mock_data(self):
        x = np.linspace(0, 99, 100)
        data_tmp = 10 * gauss1D(x, 50, 10, 1) + 1 * np.random.rand((100))
        self.ind_data += 1
        self.data_mock = DataToExport('mocktcp',
                                      data=[DataRaw('mock',
                                                    data=[np.atleast_1d(np.roll(data_tmp, self.ind_data)[0])])])

    def set_1D_Mock_data(self):
        x = np.linspace(0, 99, 100)
        data_tmp = 10 * gauss1D(x, 50, 10, 1) + 1 * np.random.rand((100))
        self.ind_data += 1
        self.data_mock = DataToExport('mocktcp', data=[DataRaw('mock', data=[np.roll(data_tmp, self.ind_data)])])

    def set_2D_Mock_data(self):
        self.x_axis = np.linspace(0, 50, 50, endpoint=False)
        self.y_axis = np.linspace(0, 30, 30, endpoint=False)
        data_tmp = 10 * gauss2D(self.x_axis, 20, 10,
                                      self.y_axis, 15, 7, 1) + 2 * np.random.rand(len(self.y_axis), len(self.x_axis))
        self.data_mock = DataToExport('mocktcp', data=[DataRaw('mock', data=[data_tmp])])

if __name__ == '__main__':
    prog = DAQ_Viewer_TCP_server()



# -*- coding: utf-8 -*-
"""
Created on Mon March 03 2021
author: Sebastien Weber
"""
import numpy as np


def random_step(start, stop, step):
    tmp = start
    out = np.array([tmp])
    sign = stop - start
    if step == 0:
        raise ValueError('step must be strictly positive or negative')
    if step > 0 and sign > 0:
        while tmp <= stop:
            tmp = tmp + (np.random.random() + 0.5) * step
            out = np.append(out, tmp)
    elif step < 0 and sign < 0:
        while tmp >= stop:
            tmp = tmp + (np.random.random() + 0.5) * step
            out = np.append(out, tmp)
    else:
        raise ValueError(f'the step value {step} is not of the same sign as stop - start : {sign}')
    return out[0:-1]


def linspace_this_vect(x, y=None, Npts=None):
    """
        Given a vector x, it returns a vector xlin where xlin is a
        linearised version of x on the same interval and with the same size.
        if args is provided it is a y vector and the function returns both xlin
        and ylin where ylin is a linear interpolation of y on this new xlin axis
    
        Parameters
        ----------
        x : (ndarray)
        y : (ndarray) optional
        Npts: (int) size of the linear vector (optional)

        Returns
        -------
        xlin : vector
        (ylin : vector) optional if args is provided

    """
    if not Npts:
        Npts = np.size(x)
    xlin = np.linspace(np.min(x), np.max(x), Npts)
    if y is not None:
        ylin = np.interp(xlin, x, y)
        return xlin, ylin
    else:
        return xlin

    
def find_index(x, threshold):
    """
    find_index finds the index ix such that x(ix) is the closest from threshold
    
    Parameters
    ----------
    x : vector
    threshold : list of scalar

    Returns
    -------
    out : list of 2-tuple containing ix,x[ix]
            out=[(ix0,xval0),(ix1,xval1),...]
    """

    if np.isscalar(threshold):
        threshold = [threshold]
    out = []
    for value in threshold:
        ix = int(np.argmin(np.abs(x - value)))
        out.append((ix, x[ix]))
    return out


def find_rising_edges(x, threshold):
    """find_rising_edges finds the index ix such that x(ix) is the closest from threshold and values are increasing
    
    Parameters
    ----------
    x : vector
    threshold : list of scalar

    Returns
    -------
    out : list of 2-tuple containing ix,x[ix]
            out=[(ix0,xval0),(ix1,xval1),...]
    
    """
    x_shifted = np.concatenate((x[1:], np.array((np.NaN,))))
    if np.isscalar(threshold):
        threshold = [threshold]
    out = []
    for value in threshold:
        dat = np.bitwise_and(x < value, x_shifted > value)
        ix = [ind for ind, flag in enumerate(dat) if flag]
        out.append((ix, x[ix]))
    return out


def crop_vector_to_axis(x, V, xlim):
    """crops a vector V with given x axis vector to a given xlim tuple
    
    Parameters
    ----------
    x : vector
    V : vector
    xlim: tuple containing (xmin,xmax)
    
    Returns
    -------
    x_c : vector
    V_c : vector
    """
    x1 = find_index(x, xlim[0])[0][0]
    x2 = find_index(x, xlim[1])[0][0]
    if x2 > x1:
        ixx = np.linspace(x1, x2, x2 - x1 + 1, dtype=int);
    else:
        ixx = np.linspace(x2, x1, x1 - x2 + 1, dtype=int);

    x_c = x[ixx]
    V_c = V[ixx]
    return x_c, V_c

def rescale(x, window=[0.0, 1.0]):
    """ Rescales a numpy array to the range specified by ``window``.

    Default is [0, 1].
    """
    maxx = np.max(x)
    minx = np.min(x)
    return (x - minx) / (maxx - minx) * (window[1] - window[0]) + window[0]


def marginals(data, normalize=False, axes=None):
    """ Calculates the marginals of the data array.

    axes specifies the axes of the marginals, e.g., the axes on which the
    sum is projected.

    If axis is None a list of all marginals is returned.
    """
    if axes is None:
        axes = range(data.ndim)
    axes = list(axes)
    full_axes = list(range(data.ndim))
    m = []
    for i in axes:
        # for the marginal sum over all axes except the specified one
        margin_axes = tuple(j for j in full_axes if j != i)
        m.append(np.sum(data, axis=margin_axes))
    if normalize:
        m = [rescale(mx) for mx in m]
    return tuple(m) if len(m) != 1 else m[0]


def find(x, condition, n=1):
    """ Return the index of the nth element that fulfills the condition.
    """
    search_n = 1
    for i in range(len(x)):
        if condition(x[i]):
            if search_n == n:
                return i
            search_n += 1
    return -1


def arglimit(y, threshold=1e-3, padding=0.0, normalize=True):
    """ Returns the first and last index where `y >= threshold * max(abs(y))`.
    """
    t = np.abs(y)
    if normalize:
        t = t / np.max(t)

    idx1 = find(t, lambda x: x >= threshold)
    if idx1 == -1:
        idx1 = 0
    idx2 = find(t[::-1], lambda x: x >= threshold)
    if idx2 == -1:
        idx2 = t.shape[0] - 1
    else:
        idx2 = t.shape[0] - 1 - idx2

    return (idx1, idx2)


def limit(x, y=None, threshold=1e-3, padding=0.25, extend=True):
    """ Returns the maximum x-range where the y-values are sufficiently large.

    Parameters
    ----------
    x : array_like
        The x values of the graph.
    y : array_like, optional
        The y values of the graph. If `None` the maximum range of `x` is
        used. That is only useful if `padding > 0`.
    threshold : float
        The threshold relative to the maximum of `y` of values that should be
        included in the bracket.
    padding : float
        The relative padding on each side in fractions of the bracket size.
    extend : bool, optional
        Signals if the returned range can be larger than the values in ``x``.
        Default is `True`.

    Returns
    -------
    xl, xr : float
        Lowest and biggest value of the range.

    """
    if y is None:
        x1, x2 = np.min(x), np.max(x)
        if not extend:
            return (x1, x2)
    else:
        idx1, idx2 = arglimit(y, threshold=threshold)
        x1, x2 = sorted([x[idx1], x[idx2]])

    # calculate the padding
    if padding != 0.0:
        pad = (x2 - x1) * padding
        x1 -= pad
        x2 += pad

    if not extend:
        x1 = max(x1, np.min(x))
        x2 = min(x2, np.max(x))

    return (x1, x2)


def crop_array_to_axis(x, y, M, cropbox):
    """crops an array M with given cropbox as a tuple (xmin,xmax,ymin,ymax).
    
    Parameters
    ----------
    x : vector
    y : vector
    M : 2D array
    cropbox: 4 elements tuple containing (xmin,xmax,ymin,ymax)
    
    Returns
    -------
    x_c : croped x vector
    y_c : croped y  vector
    M_c : croped 2D M array

    """
    x1 = find_index(x, cropbox[0])[0][0]
    x2 = find_index(x, cropbox[1])[0][0]
    if x2 > x1:
        ixx = np.linspace(x1, x2, x2 - x1 + 1, dtype=int)
    else:
        ixx = np.linspace(x2, x1, x1 - x2 + 1, dtype=int)

    y1 = find_index(y, cropbox[2])[0][0]
    y2 = find_index(y, cropbox[3])[0][0]
    if y2 > y1:
        iyy = np.linspace(y1, y2, y2 - y1 + 1, dtype=int)
    else:
        iyy = np.linspace(y2, y1, y1 - y2 + 1, dtype=int)

    x_c = x[ixx]
    y_c = y[iyy]

    M_c = M[iyy[0]:iyy[-1] + 1, ixx[0]:ixx[-1] + 1]
    return x_c, y_c, M_c


def interp1D(x, M, xlin, axis=1):
    """
    same as numpy interp function but works on 2D array
    you have to specify the axis over which to do the interpolation
    kwargs refers to the numpy interp kwargs
    returns both xlin and the new 2D array Minterp
    """
    if axis == 0:
        Minterp = np.zeros((np.size(xlin), np.size(M, axis=1)))
        indexes = np.arange(0, np.size(M, axis=1))
        for ind in indexes:
            #             print(ind)
            Minterp[:, ind] = np.interp(xlin, x, M[:, ind])
    else:
        Minterp = np.zeros((np.size(M, axis=0), np.size(xlin)))
        indexes = np.arange(0, np.size(M, axis=0))
        for ind in indexes:
            Minterp[ind, :] = np.interp(xlin, x, M[ind, :])
    return Minterp


def linspace_this_image(x, M, axis=1, Npts=None):
    """
    Given a vector x and a 2D array M, it returns an array vector xlin where xlin is a
    linearised version of x on the same interval and with the same size. it returns as well
    a 2D array Minterp interpolated on the new xlin vector along the specified axis.

    Parameters
    ----------
    x : (vector)
    M : (2D array)
    axis : (int)
    Npts: (int) size of the linear vector (optional)

    Returns
    -------
    xlin : vector
    Minterp : 2D array
    """
    xlin = linspace_this_vect(x, Npts=Npts)
    Minterp = interp1D(x, M, xlin, axis=axis)

    return xlin, Minterp


def max_ind(x, axis=None):
    """returns the max value in a vector or array and its index (in a tuple)

    Parameters
    ----------
    x : vector
    
    axis : optional dimension aginst which to normalise
      
    Returns
    -------
    ind_max : index of the maximum value
    
    max_val : maximum value
    """
    ind_max = np.argmax(x, axis=axis)
    max_val = np.max(x, axis=axis)
    return ind_max, max_val


def min_ind(x, axis=None):
    """returns the min value in a vector or array and its index (in flattened array)

    Parameters
    ----------
    x : vector
    axis : optional dimension to check the function

    Returns
    -------
    ind_min : index of the minimum value
    min_val : minimum value
    """
    ind_min = np.argmin(x, axis=axis)
    min_val = np.min(x, axis=axis)
    return ind_min, min_val


if __name__ == '__main__':                                          # pragma: no cover
    from pymodaq.utils import daq_utils as utils
    import matplotlib.pyplot as plt

    x = random_step(00, 100, 5)
    y = random_step(00, 100, 5)
    g2 = utils.gauss2D(x, 35, 15, y, 55, 20, 1)
    (xlin, g2_interp) = linspace_this_image(x, g2, axis=1, Npts=100)
    (ylin, g2_interp_both) = linspace_this_image(y, g2_interp, axis=0, Npts=100)
    plt.figure('gauss2D')
    plt.subplot(221)
    plt.pcolormesh(x, y, g2)
    plt.subplot(222)
    plt.pcolormesh(xlin, y, g2_interp)
    plt.subplot(223)
    plt.pcolormesh(xlin, ylin, g2_interp_both)

    plt.show()

    x_c, y_c, M_c = crop_array_to_axis(x, y, g2, [20, 60, 40, 80])
    plt.figure('cropped')
    plt.subplot(121)
    plt.pcolormesh(x, y, g2)
    plt.subplot(122)
    plt.pcolormesh(x_c, y_c, M_c)
    plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Automated scanning module functionalities for PyMoDAQ

Contains all objects related to the DAQScan module, to do automated scans, saving data...
"""

import sys
from collections import OrderedDict
import datetime
from typing import TYPE_CHECKING, Union

from pymodaq.utils.logger import set_logger, get_module_name
from pymodaq.utils.gui_utils.custom_app import CustomApp
from pymodaq.utils.gui_utils.dock import Dock
from pymodaq.utils.config import Config, get_set_preset_path
import pymodaq.utils.parameter.ioxml
import numpy as np
from qtpy import QtWidgets
from qtpy.QtCore import QObject, Slot, QThread, Signal, Qt
from pymodaq.utils import data as data_mod
from pymodaq.utils.gui_utils.widgets import QLED
from pymodaq.utils.gui_utils.dock import DockArea
from pymodaq.utils import daq_utils as utils
from pymodaq.utils.h5modules.h5logging import H5Logger
from pymodaq.utils.managers.modules_manager import ModulesManager
from pymodaq.utils.data import DataToExport, DataActuator, DataWithAxes


if TYPE_CHECKING:
    from pymodaq.dashboard import DashBoard



config = Config()
logger = set_logger(get_module_name(__file__))


try:
    import sqlalchemy
    from pymodaq.utils.db.db_logger.db_logger import DataBaseLogger
    is_sql = True
except Exception as e:
    is_sql = False
    logger.info('To enable logging to database install: sqalchemy and sqlalchemy_utils packages as well as the '
                'backend for your specific database, for instance psycopg2 for PostGreSQL database')

LOG_TYPES = ['None', 'H5 File']
if is_sql:
    LOG_TYPES.append('SQL DataBase')


class DAQ_Logger(CustomApp):
    """
    Main class initializing a DAQ_Logger module
    """
    command_DAQ_signal = Signal(list)
    status_signal = Signal(str)

    params = [
        {'title': 'Log Type:', 'name': 'log_type', 'type': 'str', 'value': '', 'readonly': True},
    ]

    def __init__(self, dockarea: DockArea = None, dashboard: 'DashBoard' = None):
        """

        Parameters
        ----------
        dockarea: DockArea
            instance of the modified pyqtgraph Dockarea
        dashboard: DashBoard
            instance of the pymodaq dashboard
        """

        super().__init__(dockarea, dashboard)

        self.wait_time = 1000

        self.logger_thread = None
        self.logger: Union[H5Logger, DataBaseLogger] = None
        self.setup_ui()

    def setup_actions(self):
        '''
        subclass method from ActionManager
        '''
        logger.debug('setting actions')
        self.add_action('quit', 'Quit', 'close2', "Quit program", toolbar=self.toolbar)
        self.toolbar.addSeparator()
        self.add_action('start', 'Start Logging', 'run2', "Start the logging",
                        checkable=True, toolbar=self.toolbar)
        self.add_action('stop', 'Stop', 'stop', 'Stop/pause logging',
                        checkable=False, toolbar=self.toolbar)

        log_type_combo = QtWidgets.QComboBox()
        log_type_combo.addItems(LOG_TYPES)
        log_type_combo.currentTextChanged.connect(self.set_log_type)
        self._actions['log_type'] = self.toolbar.addWidget(log_type_combo)
        self.toolbar.addSeparator()
        self.add_action('grab_all', 'Grab All', 'run_all', "Grab all selected detectors's data and actuators's value",
                        checkable=False, toolbar=self.toolbar)
        self.add_action('stop_all', 'Stop All', 'stop_all', "Stop all selected detectors and actuators",
                        checkable=False, toolbar=self.toolbar)
        self.add_action('infos', 'Log infos', 'information2', "Show log file",
                        checkable=False, toolbar=self.toolbar)

        self.set_action_enabled('start', False)
        self.set_action_enabled('stop', False)

        logger.debug('actions set')

    def setup_docks(self):
        logger.debug('setting docks')
        self.docks['detectors'] = Dock("Detectors")
        splitter = QtWidgets.QSplitter(Qt.Vertical)
        self.docks['detectors'].addWidget(splitter)
        splitter.addWidget(self.settings_tree)
        splitter.addWidget(self.modules_manager.settings_tree)
        self.dockarea.addDock(self.docks['detectors'])

        self.docks['logger_settings'] = Dock("Logger Settings")
        self.dockarea.addDock(self.docks['logger_settings'], 'right', self.docks['detectors'])

        self.statusbar.setMaximumHeight(25)
        self.status_widget = QtWidgets.QLabel('Initializing')
        self.statusbar.addPermanentWidget(self.status_widget)

        self.start_log_time = QtWidgets.QDateTimeEdit()
        self.start_log_time.setReadOnly(True)
        self.start_log_time.setToolTip('Logging started at:')
        self.statusbar.addPermanentWidget(self.start_log_time)

        self.logging_state = QLED()
        self.logging_state.setToolTip('logging status: green (running), red (idle)')
        self.logging_state.clickable = False
        self.statusbar.addPermanentWidget(self.logging_state)

    def connect_things(self):
        self.status_signal[str].connect(self.dashboard.add_status)
        self._actions['quit'].connect_to(self.quit_fun)

        self._actions['start'].connect_to(self.start_logging)
        self._actions['stop'].connect_to(self.stop_logging)
        self._actions['grab_all'].connect_to(self.start_all)
        self._actions['stop_all'].connect_to(self.stop_all)

        self._actions['infos'].connect_to(self.dashboard.show_log)

    def setup_menu(self):
        """
        """
        file_menu = self.mainwindow.menuBar().addMenu('File')
        self.affect_to('infos', file_menu)

    def value_changed(self, param):
        if param.name() == 'log_type':
            if param.value() != 'None':
                self.set_action_enabled('start', True)
                self.set_action_enabled('stop', True)
                self.set_logger(param.value())

    def set_logger(self, logger_interface):
        if self.logger is not None:
            self.logger.close()
            self.docks['logger_settings'].removeWidgets()

        if logger_interface == 'H5 File':
            self.logger = H5Logger(self.modules_manager)
        elif logger_interface == 'SQL DataBase':
            self.logger = DataBaseLogger(self.dashboard.preset_file.stem)
        else:
            return

        self.docks['logger_settings'].addWidget(self.logger.settings_tree)

    def quit_fun(self):
        """
            Quit the current instance of DAQ_scan and close on cascade move and detector modules.

            See Also
            --------
            quit_fun
        """
        try:
            self.logger.close()
        except Exception as e:
            logger.exception(str(e))

        self.dockarea.close()

    def set_continuous_save(self):
        """
            Set a continous save file using the base path located file with
            a header-name containing date as a string.

            See Also
            --------
            daq_utils.set_current_scan_path
        """
        if self.logger is not None:
            self.do_continuous_save = True
            self.logger.settings.child('N_saved').show()
            self.logger.settings.child('N_saved').setValue(0)

            settings_str = b'<All_settings>'
            settings_str += pymodaq.utils.parameter.ioxml.parameter_to_xml_string(self.dashboard.settings)
            settings_str += pymodaq.utils.parameter.ioxml.parameter_to_xml_string(
                self.dashboard.preset_manager.preset_params)
            if self.dashboard.settings.child('loaded_files', 'overshoot_file').value() != '':
                settings_str += pymodaq.utils.parameter.ioxml.parameter_to_xml_string(
                    self.dashboard.overshoot_manager.overshoot_params)
            if self.dashboard.settings.child('loaded_files', 'roi_file').value() != '':
                settings_str += pymodaq.utils.parameter.ioxml.parameter_to_xml_string(
                    self.dashboard.roi_saver.roi_presets)
            settings_str += pymodaq.utils.parameter.ioxml.parameter_to_xml_string(self.settings)
            settings_str += pymodaq.utils.parameter.ioxml.parameter_to_xml_string(self.logger.settings)
            settings_str += b'</All_settings>'

            if not self.logger.init_logger(settings_str):
                return False
            logger.addHandler(self.logger.get_handler())
            return True
        else:
            return False

    def set_logging(self):
        """

        """
        status = self.set_continuous_save()
        if status:
            modules_log = self.modules_manager.detectors_all + self.modules_manager.actuators_all
            if modules_log != []:
                # check if the modules are initialized
                for module in modules_log:
                    if not module.initialized_state:
                        logger.error(f'module {module.title} is not initialized')
                        return False

                # create the detectors in the chosen logger
                for mod in modules_log:
                    settings_str = b'<All_settings>'
                    settings_str += pymodaq.utils.parameter.ioxml.parameter_to_xml_string(mod.settings)

                    if mod.module_type == 'DAQ_Viewer':
                        for viewer in mod.ui.viewers:
                            if hasattr(viewer, 'roi_manager'):
                                settings_str += pymodaq.utils.parameter.ioxml.parameter_to_xml_string(
                                    viewer.roi_manager.settings)
                    settings_str += b'</All_settings>'
                    if mod.module_type == 'DAQ_Viewer':
                        self.logger.add_detector(mod.title, settings_str)
                    elif mod.module_type == 'DAQ_Move':
                        self.logger.add_actuator(mod.title, settings_str)

                self._actions['start'].setEnabled(True)
                self._actions['stop'].setEnabled(True)
                return True
            else:
                self.update_status('Cannot start logging... No detectors selected')
                self._actions['start'].setEnabled(False)
                self._actions['stop'].setEnabled(True)
                return False

        else:
            self.update_status('Cannot start logging... check connections')
            self._actions['start'].setEnabled(False)
            self._actions['stop'].setEnabled(True)
            return False

    def start_all(self):
        for det in self.modules_manager.detectors:
            det.grab()
        for act in self.modules_manager.actuators:
            act.grab()

    def stop_all(self):
        for det in self.modules_manager.detectors:
            det.stop_grab()
        for act in self.modules_manager.actuators:
            act.stop_grab()

    def set_log_type(self, log_type):
        self.settings.child('log_type').setValue(log_type)

    def start_logging(self):
        """
            Start a logging.
        """
        self.status_widget.setText('Starting logging')

        self.overshoot = False
        res = self.set_logging()

        # mandatory to deal with multithreads
        if self.logger_thread is not None:
            self.command_DAQ_signal.disconnect()
            if self.logger_thread.isRunning():
                self.logger_thread.exit()
                while not self.logger_thread.isFinished():
                    QThread.msleep(100)
                self.logger_thread = None

        self.logger_thread = QThread()

        log_acquisition = DAQ_Logging(self.settings, self.logger, self.modules_manager)

        log_acquisition.moveToThread(self.logger_thread)

        self.command_DAQ_signal[list].connect(log_acquisition.queue_command)
        log_acquisition.status_sig[list].connect(self.thread_status)

        self.logger_thread.log_acquisition = log_acquisition
        self.logger_thread.start()

        self._actions['start'].setEnabled(False)
        QtWidgets.QApplication.processEvents()
        self.logging_state.set_as_false()

        self.command_DAQ_signal.emit(["start_logging"])
        self.status_widget.setText('Running acquisition')

    def stop_logging(self):
        """
            Emit the command_DAQ signal "stop_acquisiion".

            See Also
            --------
            set_ini_positions
        """
        preset_items_det = self.modules_manager.detectors
        for det in preset_items_det:
            det.stop()
        self.command_DAQ_signal.emit(["stop_acquisition"])

        if not self.dashboard.overshoot:
            status = 'Data Logging has been stopped by user'
        else:
            status = 'Data Logging has been stopped due to overshoot'

        self.update_status(status)
        self._actions['start'].setEnabled(True)

    @Slot(list)
    def thread_status(self, status):  # general function to get datas/infos from all threads back to the main
        """
            | General function to get datas/infos from all threads back to the main.
            |

            Switch the status with :
                * *"Update status"* : Update the status bar with the status attribute txt message
                * *"Update_scan_index"* : Set the value of the User Interface - indice_scan_sb attribute.
                * *"Scan_done"* : Save the scan and init the positions
                * *"Timeout"* : Set the "Timeout occured" in the User Interface-log message

            See Also
            --------
            update_status, save_scan, set_ini_positions
        """
        if status[0] == "Update_Status":
            self.update_status(status[1], wait_time=self.wait_time)

        elif status[0] == "Timeout":
            self.status_widget.setText('Timeout occurred')

    def update_status(self, txt, wait_time=0):
        """
            Show the txt message in the status bar with a delay of wait_time ms.

            =============== =========== =======================
            **Parameters**    **Type**    **Description**
            *txt*             string      The message to show
            *wait_time*       int         the delay of showing
            *log_type*        string      the type of the log
            =============== =========== =======================
        """
        try:
            self.statusbar.showMessage(txt, wait_time)
            logger.info(txt)
        except Exception as e:
            logger.exception(str(e))


class DAQ_Logging(QObject):
    """

    """

    scan_data_tmp = Signal(OrderedDict)
    status_sig = Signal(list)

    def __init__(self, settings=None, logger=None, modules_manager: ModulesManager = None):

        """
            DAQ_Logging deal with the acquisition part of daq_scan.

            See Also
            --------
            custom_tree.parameter_to_xml_string
        """
        
        super(QObject, self).__init__()

        self.stop_logging_flag = False
        self.settings = settings
        self.ini_time = 0
        self.ind_log = 0
        self.modules_manager: ModulesManager = modules_manager
        self.modules_manager.detectors_changed.connect(self.update_connect_detectors)
        self.modules_manager.actuators_changed.connect(self.update_connect_actuators)
        self.data_logger = logger

    @Slot(list)
    def queue_command(self, command):
        """
            Treat the queue of commands from the current command to act, between :
                * *start_logging*
                * *stop_acquisition*
                * *set_ini_position*
                * *move_stages*

            =============== ============== =========================
            **Parameters**    **Type**      **Description**
            command           string list   the command string list
            =============== ============== =========================

            See Also
            --------
            start_logging, set_ini_positions, move_stages
        """
        if command[0] == "start_logging":
            self.start_logging()

        elif command[0] == "stop_acquisition":
            self.stop_scan_flag = True
            self.stop_logging()

    def do_save_continuous(self, data):
        """

        """
        try:
            self.data_logger.add_data(data)
        except Exception as e:
            logger.exception(str(e))

    def format_actuators_data(self, data_act: DataActuator):
        data = data_mod.DataToExport(name=data_act.name, data=[data_act])
        self.do_save_continuous(data)

    def connect_actuators(self, connect=True):
        """Connect actuators to DAQ_Logging do_save_continuous method

        Parameters
        ----------
        connect: bool
            If True make the connection else disconnect
        """
        self.modules_manager.connect_actuators(connect=connect, slot=self.format_actuators_data, signal='current_value')

    def update_connect_actuators(self):
        try:
            self.connect_actuators(False)
        except:
            pass
        self.connect_actuators()

    def connect_detectors(self, connect=True):
        """Connect detectors to DAQ_Logging do_save_continuous method

        Parameters
        ----------
        connect: bool
            If True make the connection else disconnect
        """
        self.modules_manager.connect_detectors(connect=connect, slot=self.do_save_continuous)

    def update_connect_detectors(self):
        try:
            self.connect_detectors(False)
        except :
            pass
        self.connect_detectors()

    def stop_logging(self):
        try:
            self.connect_detectors(connect=False)
            self.connect_actuators(connect=False)
        except Exception as e:
            logger.exception(str(e))

        if self.stop_logging_flag:
            status = 'Data Acquisition has been stopped by user'
            self.status_sig.emit(["Update_Status", status])
        if self.data_logger is not None:
            self.data_logger.stop_logger()

    def start_logging(self):
        try:
            self.connect_detectors()
            self.connect_actuators()
            self.stop_logging_flag = False
            self.status_sig.emit(["Update_Status", "Acquisition has started"])

        except Exception as e:
            logger.exception(str(e))


def main():
    from pymodaq.dashboard import DashBoard
    from pathlib import Path
    from pymodaq.utils.gui_utils.dock import DockArea

    config = Config()
    app = QtWidgets.QApplication(sys.argv)
    if config('style', 'darkstyle'):
        import qdarkstyle
        app.setStyleSheet(qdarkstyle.load_stylesheet())

    win = QtWidgets.QMainWindow()
    area = DockArea()
    win.setCentralWidget(area)
    win.resize(1000, 500)
    win.setWindowTitle('PyMoDAQ Dashboard')

    # win.setVisible(False)
    prog = DashBoard(area)
    file = Path(get_set_preset_path()).joinpath(f"{config('presets', 'default_preset_for_logger')}.xml")
    if file.exists():
        prog.set_preset_mode(file)
        prog.load_log_module()
    else:
        msgBox = QtWidgets.QMessageBox()
        msgBox.setText(f"The default file specified in the configuration file does not exists!\n"
                       f"{file}\n"
                       f"Impossible to load the Logger Module")
        msgBox.setStandardButtons(msgBox.Ok)
        ret = msgBox.exec()

    sys.exit(app.exec_())


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
Created the 05/12/2022

@author: Sebastien Weber
"""
from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import Callable, Union, List, Tuple, TYPE_CHECKING


import numpy as np
from qtpy import QtWidgets

from pymodaq.utils.managers.parameter_manager import ParameterManager, Parameter
from pymodaq.utils.parameter.utils import get_param_path, iter_children_params
from pymodaq.utils.factory import ObjectFactory
from pymodaq.utils.logger import set_logger, get_module_name
from pymodaq.utils.data import Axis, DataDistribution
from pymodaq.utils.abstract import abstract_attribute
from pymodaq.utils import math_utils as mutils
from pymodaq.utils import config as configmod
from pymodaq.utils.scanner.scan_config import ScanConfig

if TYPE_CHECKING:
    from pymodaq.control_modules.daq_move import DAQ_Move
    from pymodaq.utils.plotting.scan_selector import Selector


logger = set_logger(get_module_name(__file__))
config = configmod.Config()


class ScanParameterManager(ParameterManager):
    settings_name = 'scanner_settings'

    def __init__(self):
        super().__init__()
        self.settings_tree.header().setVisible(True)
        self.settings_tree.header().setSectionResizeMode(QtWidgets.QHeaderView.ResizeMode.Interactive)
        self.settings_tree.header().setMinimumSectionSize(150)
        self.settings_tree.setMinimumHeight(150)


class ScannerBase(ScanParameterManager, metaclass=ABCMeta):
    """Abstract class for all Scanners

    Attributes
    ----------
    scan_type: str
        String defining the main identifier
    scan_subtype: str
        String defining the second identifier
    params: List[dict]
        list specifying the scanner set of parameters to properly configure all the scan steps
    positions: np.ndarray
        ndarray of all positions. First dimension is number of actuators, second is positions of a given actuator at
        each step
    axes_unique: List[np.ndarray]
        list of ndarrays representing unique values of steps for each actuator in the scan
    axes_indexes: np.ndarray
        ndarray of indexes from axes_unique for each value in positions
    n_steps: int
        Number of scan steps. Equal to the second dimension of positions
    n_axes: int
        Number of actuators/scan axes. Equal to the first dimension of positions
    """
    scan_type: str = abstract_attribute()
    scan_subtype: str = abstract_attribute()

    params: List[dict] = abstract_attribute()
    axes_unique: List[np.ndarray] = abstract_attribute()
    axes_indexes: np.ndarray = abstract_attribute()
    n_steps: int = abstract_attribute()
    n_axes: int = abstract_attribute()
    distribution: DataDistribution = abstract_attribute()
    save_settings = True

    def __init__(self, actuators: List[DAQ_Move] = None):
        super().__init__()
        self.positions: np.ndarray = None
        self.n_steps = 1
        self._actuators: List[DAQ_Move] = None

        self.actuators = actuators

        self.config = ScanConfig()

        self.set_settings_titles()
        self.set_settings_values()

        if self.check_steps():
            self.set_scan()


    def set_settings_titles(self):
        """Update the settings accordingly with the selected actuators"""
        ...

    def set_settings_values(self, param: Parameter = None):
        if param is None:
            param = self.settings
        for child in param.children():
            if len(child.children()) == 0:  #means it is not a group parameter
                path = self.actuators_name
                path_scan = [self.scan_type, self.scan_subtype]
                path_param = get_param_path(child)[1:]
                path.extend(path_scan)
                path.extend(path_param)
                try:
                    child.setValue(self.config(*path))  # first try to load the config including the actuators name
                except configmod.ConfigError as e:
                    try:
                        path = path_scan
                        path.extend(path_param)
                        child.setValue(self.config(*path))   # then without the actuators name
                    except Exception as e:
                        pass
            else:
                self.set_settings_values(child)

    @property
    def actuators(self) -> List[DAQ_Move]:
        return self._actuators

    @actuators.setter
    def actuators(self, actuators: List[DAQ_Move]):
        self._actuators = actuators

    @property
    def actuators_name(self) -> List[str]:
        return [act.title for act in self.actuators]

    def check_steps(self):
        steps_limit = config('scan', 'steps_limit')
        n_steps = self.evaluate_steps()
        return n_steps <= steps_limit

    def __call__(self, **kwargs):
        return self(**kwargs)

    @abstractmethod
    def set_scan(self):
        """To be reimplemented. Calculations of all mandatory attributes from the settings"""
        ...

    @abstractmethod
    def get_nav_axes(self) -> List[Axis]:
        """To be reimplemented. Calculations of all navigation axes from attributes"""
        ...

    @abstractmethod
    def get_scan_shape(self) -> Tuple[int]:
        """To be reimplemented. Calculations of all the final shape of the scan"""
        ...

    @abstractmethod
    def get_indexes_from_scan_index(self, scan_index: int) -> Tuple[int]:
        """To be reimplemented. Calculations of indexes within the scan"""
        ...

    def get_info_from_positions(self, positions: np.ndarray):
        """Set mandatory attributes from a ndarray of positions"""
        if positions is not None:
            if len(positions.shape) == 1:
                positions = np.expand_dims(positions, 1)
            axes_unique = []
            for ax in positions.T:
                axes_unique.append(np.unique(ax))
            axes_indexes = np.zeros_like(positions, dtype=int)
            for ind in range(positions.shape[0]):
                for ind_pos, pos in enumerate(positions[ind]):
                    axes_indexes[ind, ind_pos] = mutils.find_index(axes_unique[ind_pos], pos)[0][0]

            self.n_axes = len(axes_unique)
            self.axes_unique = axes_unique
            self.axes_indexes = axes_indexes
            self.positions = positions
            self.n_steps = positions.shape[0]

    @abstractmethod
    def evaluate_steps(self):
        """To be reimplemented. Quick evaluation of the number of steps to stop the calculation if the evaluation os above the
        configured limit
        """
        ...

    def update_model(self):
        """Method to reimplement and use for scanners using table_view types Parameters to set and apply the underlying
        model

        See Also
        --------
        SequentialScanner, TabularScanner, pymodaq.utils.parameter.pymodaq_ptypes.tableview
        """
        ...

    def save_scan_parameters(self):
        if self.save_settings:
            path_actuators = self.actuators_name
            path_scan = [self.scan_type, self.scan_subtype]
            path_actuators.extend(path_scan)
            for param in iter_children_params(self.settings, []):
                path = path_actuators[:]
                path_param = get_param_path(param)[1:]
                path.extend(path_param)
                try:
                    self.config[tuple(path)] = param.value()
                except Exception as e:
                    pass
            self.config.save()

    @abstractmethod
    def update_from_scan_selector(self, scan_selector: Selector):
        """To be reimplemented. Process the Selector object to set the Scanner settings

        See Also
        --------
        Selector, ScanSelector
        """
        ...


class ScannerFactory(ObjectFactory):
    """Factory class registering and storing Scanners"""

    @classmethod
    def register(cls) -> Callable:
        """ To be used as a decorator

        Register in the class registry a new scanner class using its 2 identifiers: scan_type and scan_sub_type
        """
        def inner_wrapper(wrapped_class: ScannerBase) -> Callable:
            if cls.__name__ not in cls._builders:
                cls._builders[cls.__name__] = {}
            key = wrapped_class.scan_type
            sub_key = wrapped_class.scan_subtype

            if key not in cls._builders[cls.__name__]:
                cls._builders[cls.__name__][key] = {}
            if sub_key not in cls._builders[cls.__name__][key]:
                cls._builders[cls.__name__][key][sub_key] = wrapped_class
            else:
                logger.warning(f'The {cls.__name__}/{key}/{sub_key} builder is already registered. Replacing it')
            return wrapped_class

        return inner_wrapper

    @classmethod
    def create(cls, key, sub_key, **kwargs) -> ScannerBase:
        builder = cls._builders[cls.__name__].get(key).get(sub_key)
        if not builder:
            raise ValueError(key)
        return builder(**kwargs)

    def get(self, scan_type, scan_sub_type, **kwargs):
        return self.create(scan_type, scan_sub_type, **kwargs)

    def scan_types(self) -> List[str]:
        """Returns the list of scan types, main identifier of a given scanner"""
        return sorted(list(self.builders[self.__class__.__name__].keys()))

    def scan_sub_types(self, scan_type: str) -> List[str]:
        """Returns the list of scan subtypes, second identifier of a given scanner of type scan_type"""
        return list(self.builders[self.__class__.__name__][scan_type].keys())




# -*- coding: utf-8 -*-
"""
Created the 05/09/2022

@author: Sebastien Weber
"""


from typing import List
import sys

from qtpy import QtWidgets
from qtpy.QtCore import Signal
from qtpy.QtWidgets import QVBoxLayout,  QWidget, QComboBox
from pymodaq.utils.daq_utils import ThreadCommand
from pymodaq.control_modules.utils import ControlModuleUI

from pymodaq.utils.gui_utils.widgets import PushButtonIcon, LabelWithFont, QLED
from pymodaq.utils.gui_utils import Dock
from pymodaq.utils.config import Config
from pymodaq.control_modules.utils import DET_TYPES, DAQTypesEnum
from pymodaq.utils.plotting.data_viewers.viewer import ViewerFactory, ViewersEnum, ViewerDispatcher
from pymodaq.utils.enums import enum_checker


viewer_factory = ViewerFactory()
config = Config()


class DAQ_Viewer_UI(ControlModuleUI, ViewerDispatcher):
    """DAQ_Viewer user interface.

    This class manages the UI and emit dedicated signals depending on actions from the user

    Attributes
    ----------
    command_sig: Signal[Threadcommand]
        This signal is emitted whenever some actions done by the user has to be
        applied on the main module. Possible commands are:
            * init
            * quit
            * grab
            * snap
            * stop
            * show_log
            * detector_changed
            * daq_type_changed
            * save_current
            * save_new

    Methods
    -------
    display_value(value: float)
        Update the display of the actuator's value on the UI
    do_init()
        Programmatic init

    See Also
    --------
    pymodaq.utils.daq_utils.ThreadCommand
    """

    command_sig = Signal(ThreadCommand)

    def __init__(self, parent, title="DAQ_Viewer", daq_type='DAQ2D', dock_settings=None, dock_viewer=None):
        ControlModuleUI.__init__(self, parent)
        ViewerDispatcher.__init__(self, self.dockarea, title=title, next_to_dock=dock_settings)

        self.title = title

        self._detector_widget = None
        self._settings_widget = None
        self._info_detector = None
        self._daq_types_combo = None
        self._detectors_combo = None
        self._ini_det_pb = None
        self._ini_state_led = None
        self._do_bkg_cb = None
        self._take_bkg_pb = None
        self._settings_dock = dock_settings
        self.setup_docks()

        daq_type = enum_checker(DAQTypesEnum, daq_type)
        self.daq_types = daq_type.names()  # init the combobox through the daq_types attribute
        self.daq_type = daq_type

        self.detectors = [det['name'] for det in DET_TYPES[self.daq_type.name]]
        self.setup_actions()  # see ActionManager MixIn class
        self.add_viewer(self.daq_type.to_viewer_type(), dock_viewer=dock_viewer)
        self.connect_things()

        self._enable_grab_buttons(False)
        self._detector_widget.setVisible(False)
        self._settings_widget.setVisible(False)

    @property
    def detector(self):
        return self._detectors_combo.currentText()

    @detector.setter
    def detector(self, det_name: str):
        self._detectors_combo.setCurrentText(det_name)
    @property
    def detectors(self):
        return [self._detectors_combo.itemText(ind) for ind in range(self._detectors_combo.count())]

    @detectors.setter
    def detectors(self, detectors: List[str]):
        #self._detectors_combo.currentTextChanged.disconnect()
        self._detectors_combo.clear()
        self._detectors_combo.addItems(detectors)
        #self._detectors_combo.currentTextChanged.connect(
        #    lambda mod: self.command_sig.emit(ThreadCommand('detector_changed', mod)))
        #self.detector = detectors[0]

    @property
    def daq_type(self):
        return DAQTypesEnum[self._daq_types_combo.currentText()]

    @daq_type.setter
    def daq_type(self, dtype: DAQTypesEnum):
        dtype = enum_checker(DAQTypesEnum, dtype)
        self._daq_types_combo.setCurrentText(dtype.name)

    @property
    def daq_types(self):
        return self.daq_type.names()

    @daq_types.setter
    def daq_types(self, dtypes: List[str]):
        self._daq_types_combo.clear()
        self._daq_types_combo.addItems(dtypes)
        self.daq_type = DAQTypesEnum[dtypes[0]]

    def close(self):
        for dock in self.viewer_docks:
            dock.close()
        self._settings_dock.close()

    def setup_docks(self):
        if self._settings_dock is None:
            self._settings_dock = Dock(self.title + "_Settings", size=(150, 250))
            self.dockarea.addDock(self._settings_dock)

        widget = QWidget()
        widget.setLayout(QVBoxLayout())
        #widget.layout().setSizeConstraint(QHBoxLayout.SetFixedSize)
        widget.layout().setContentsMargins(2, 2, 2, 2)
        self._settings_dock.addWidget(widget)

        info_ui = QWidget()
        self._detector_widget = QWidget()
        self._settings_widget = QWidget()
        self._settings_widget.setLayout(QtWidgets.QVBoxLayout())
        bkg_widget = QWidget()
        bkg_widget.setLayout(QtWidgets.QHBoxLayout())

        widget.layout().addWidget(info_ui)
        widget.layout().addWidget(self.toolbar)
        widget.layout().addWidget(self._detector_widget)
        widget.layout().addWidget(self._settings_widget)
        widget.layout().addStretch(0)

        info_ui.setLayout(QtWidgets.QHBoxLayout())
        info_ui.layout().addWidget(LabelWithFont(self.title, font_name="Tahoma", font_size=14, isbold=True,
                                                 isitalic=True))
        self._info_detector = LabelWithFont('', font_name="Tahoma", font_size=8, isbold=True, isitalic=True)
        info_ui.layout().addWidget(self._info_detector)

        self._detector_widget.setLayout(QtWidgets.QGridLayout())
        self._daq_types_combo = QComboBox()
        self._daq_types_combo.setSizeAdjustPolicy(QComboBox.AdjustToContents)
        self._detectors_combo = QComboBox()
        self._detectors_combo.setSizeAdjustPolicy(QComboBox.AdjustToContents)
        self._ini_det_pb = PushButtonIcon('ini', 'Init. Detector', True, 'Initialize selected detector')
        self._ini_state_led = QLED(readonly=True)
        self._do_bkg_cb = QtWidgets.QCheckBox('Do Bkg')
        self._take_bkg_pb = QtWidgets.QPushButton('Take Bkg')
        self._take_bkg_pb.setChecked(False)

        self._detector_widget.layout().addWidget(LabelWithFont('DAQ type:'), 0, 0)
        self._detector_widget.layout().addWidget(self._daq_types_combo, 0, 1)
        self._detector_widget.layout().addWidget(LabelWithFont('Detector:'), 1, 0)
        self._detector_widget.layout().addWidget(self._detectors_combo, 1, 1)
        self._detector_widget.layout().addWidget(self._ini_det_pb, 0, 2)
        self._detector_widget.layout().addWidget(self._ini_state_led, 1, 2)
        self._detector_widget.layout().addWidget(bkg_widget, 2, 0, 1, 3)

        bkg_widget.layout().addWidget(self._do_bkg_cb)
        bkg_widget.layout().addWidget(self._take_bkg_pb)

        self.statusbar = QtWidgets.QStatusBar()
        self.statusbar.setMaximumHeight(30)
        widget.layout().addWidget(self.statusbar)

    def add_setting_tree(self, tree):
        self._settings_widget.layout().addWidget(tree)

    def setup_actions(self):
        self.add_action('grab', 'Grab', 'run2', "Grab data from the detector", checkable=True)
        self.add_action('snap', 'Snap', 'snap', "Take a snapshot from the detector")
        self.add_action('stop', 'Stop', 'stop', "Stop grabing")
        self.add_action('save_current', 'Save Current Data', 'SaveAs', "Save Current Data")
        self.add_action('save_new', 'Save New Data', 'Snap&Save', "Save New Data")
        self.add_action('open', 'Load Data', 'Open', "Load Saved Data")

        self.add_action('show_controls', 'Show Controls', 'Settings', "Show Controls to set DAQ and Detector type",
                        checkable=True)
        self.add_action('show_settings', 'Show Settings', 'tree', "Show Settings", checkable=True)

        self.add_action('quit', 'Quit the module', 'close2')
        self.add_action('show_config', 'Show Config', 'Settings', "Show PyMoDAQ Config", checkable=False,
                        toolbar=self.toolbar)
        self.add_action('log', 'Show Log file', 'information2')

        self._data_ready_led = QLED(readonly=True)
        self.toolbar.addWidget(self._data_ready_led)

    def connect_things(self):
        self.connect_action('show_controls', lambda show: self._detector_widget.setVisible(show))
        self.connect_action('show_settings', lambda show: self._settings_widget.setVisible(show))
        self.connect_action('quit', lambda: self.command_sig.emit(ThreadCommand('quit', )))
        self.connect_action('show_config', lambda: self.command_sig.emit(ThreadCommand('show_config', )))

        self.connect_action('log', lambda: self.command_sig.emit(ThreadCommand('show_log', )))
        self.connect_action('stop', lambda: self.command_sig.emit(ThreadCommand('stop', )))
        self.connect_action('stop', lambda: self.get_action('grab').setChecked(False))
        self.connect_action('stop', lambda: self._enable_ini_buttons(True))
        self.connect_action('stop', lambda: self._settings_widget.setEnabled(True))

        self.connect_action('grab', self._grab)
        self.connect_action('snap', lambda: self.command_sig.emit(ThreadCommand('snap', )))

        self.connect_action('save_current', lambda: self.command_sig.emit(ThreadCommand('save_current', )))
        self.connect_action('save_new', lambda: self.command_sig.emit(ThreadCommand('save_new', )))
        self.connect_action('open', lambda: self.command_sig.emit(ThreadCommand('open', )))

        self._ini_det_pb.clicked.connect(self.send_init)

        self._detectors_combo.currentTextChanged.connect(
            lambda mod: self.command_sig.emit(ThreadCommand('detector_changed', mod)))
        self._daq_types_combo.currentTextChanged.connect(self._daq_type_changed)


        self._do_bkg_cb.clicked.connect(lambda checked: self.command_sig.emit(ThreadCommand('do_bkg', checked)))
        self._take_bkg_pb.clicked.connect(lambda: self.command_sig.emit(ThreadCommand('take_bkg')))

    def update_viewers(self, viewers_type: List[ViewersEnum]):
        super().update_viewers(viewers_type)
        self.command_sig.emit(ThreadCommand('viewers_changed', attribute=dict(viewer_types=self.viewer_types,
                                                                              viewers=self.viewers)))

    @property
    def data_ready(self):
        return self._data_ready_led.get_state()

    @data_ready.setter
    def data_ready(self, status):
        self._data_ready_led.set_as(status)

    def _daq_type_changed(self, daq_type: DAQTypesEnum):
        try:
            daq_type = enum_checker(DAQTypesEnum, daq_type)

            self.command_sig.emit(ThreadCommand('daq_type_changed', daq_type))
            if self.viewer_types != [daq_type.to_viewer_type()]:
                self.update_viewers([daq_type.to_viewer_type()])
        except ValueError as e:
            pass

    def show_settings(self, show=True):
        if (self.is_action_checked('show_settings') and not show) or \
                (not self.is_action_checked('show_settings') and show):
            self.get_action('show_settings').trigger()
            
    def show_controls(self, show=True):
        if (self.is_action_checked('show_controls') and not show) or \
                (not self.is_action_checked('show_controls') and show):
            self.get_action('show_controls').trigger()

    def _grab(self):
        """Slot from the *grab* action"""
        self.command_sig.emit(ThreadCommand('grab', attribute=self.is_action_checked('grab')))
        self._enable_ini_buttons(not self.is_action_checked('grab'))
        if not self.config('viewer', 'allow_settings_edition'):
            self._settings_widget.setEnabled(not self.is_action_checked('grab'))

    def do_init(self, do_init=True):
        """Programmatically press the Init button
        API entry
        Parameters
        ----------
        do_init: bool
            will fire the Init button depending on the argument value and the button check state
        """
        if do_init is not self._ini_det_pb.isChecked():
            self._ini_det_pb.click()

    def do_grab(self, do_grab=True):
        """Programmatically press the Grab button
        API entry
        Parameters
        ----------
        do_grab: bool
            will fire the Init button depending on the argument value and the button check state
        """
        if (do_grab and not self.is_action_checked('grab')) or ((not do_grab) and self.is_action_checked('grab')):
            self.get_action('grab').trigger()

    def do_snap(self):
        """Programmatically press the Snap button
        API entry
        """
        self.get_action('snap').trigger()

    def do_stop(self):
        """Programmatically press the Stop button
        API entry
        """
        self.get_action('stop').trigger()
        if self.is_action_checked('grab'):
            self.get_action('grab').trigger()

    def send_init(self, checked: bool):
        self._enable_detchoices(not checked)
        self.command_sig.emit(ThreadCommand('init', [checked,
                                                     self._daq_types_combo.currentText(),
                                                     self._detectors_combo.currentText()]))

    def _enable_detchoices(self, enable=True):
        self._detectors_combo.setEnabled(enable)
        self._daq_types_combo.setEnabled(enable)

    @property
    def detector_init(self):
        """bool: the status of the init LED."""
        return self._ini_state_led.get_state()

    @detector_init.setter
    def detector_init(self, status):
        if status:
            self._info_detector.setText(f'{self.daq_type.name} : {self.detector}')
        else:
            self._info_detector.setText('')
        self._ini_state_led.set_as(status)
        self._enable_grab_buttons(status)

    def _enable_grab_buttons(self, status):
        self.get_action('grab').setEnabled(status)
        self.get_action('snap').setEnabled(status)
        self.get_action('stop').setEnabled(status)
        self.get_action('save_current').setEnabled(status)
        self.get_action('save_new').setEnabled(status)

    def _enable_ini_buttons(self, status):
        self._ini_det_pb.setEnabled(status)
        self.get_action('quit').setEnabled(status)


def main(init_qt=True):
    from pymodaq.utils.gui_utils.dock import DockArea
    from pymodaq.utils.parameter import ParameterTree, Parameter
    from pymodaq.control_modules.viewer_utility_classes import params as daq_viewer_params

    if init_qt:  # used for the test suite
        app = QtWidgets.QApplication(sys.argv)

    param = Parameter.create(name='settings', type='group', children=daq_viewer_params)
    tree = ParameterTree()
    tree.setParameters(param, showTop=False)

    dockarea = DockArea()
    prog = DAQ_Viewer_UI(dockarea)
    dockarea.show()

    def print_command_sig(cmd_sig):
        print(cmd_sig)
        prog.display_status(str(cmd_sig))
        if cmd_sig.command == 'init':
            prog._enable_grab_buttons(cmd_sig.attribute[0])
            prog.detector_init = cmd_sig.attribute[0]

    # prog.detectors = detectors
    prog.command_sig.connect(print_command_sig)

    prog.add_setting_tree(tree)

    prog.update_viewers([ViewersEnum['Viewer0D'], ViewersEnum['Viewer1D'], ViewersEnum['Viewer2D']])

    if init_qt:
        sys.exit(app.exec_())
    return prog, dockarea


if __name__ == '__main__':
    main()
